{-----------------------------------------------------------------------------
 Unit Name: pbCocoAutomaton
 Author:     Originally Pat Terry, ported to Delphi by Pat Connors,
 - notes from Pat Terry's CRA.PAS:
  CRA     Automaton and Scanner Generation
   ===     ================================

  (1) ConvertToStates translates a top-down graph into a NFA.
      MatchDFA tries to match literal strings against the DFA
  (2) MakeDeterministic converts the NFA into a DFA
  (3) WriteScanner generates the scanner source file

-----------------------------------------------------------------------------}


unit pbCocoAutomaton;

interface

uses
  Windows, Messages, SysUtils, Classes,
  pbCocoTable, pbCocoScanner, pbCocoSets, pbStrLib,
  Dialogs, JclSysUtils, pbCocoConstants;


type
  TPutSProc = procedure(S: string) of object;
  TComment = ^TCommentNode;
  STRING2 = string[2];
  TCommentNode = record (* info about a TComment syntax *)
    start, stop: STRING2;
    nested: boolean;
    Next: TComment;
  end;
  TMelted = ^TMeltedNode;
  TMeltedNode = record (* info about TMelted states *)
    sset: TCRTSet; (* set of old states *)
    theState: Integer; (* new state *)
    Next: TMelted;
  end;

const
  //maxStates = 500;
  maxStates = 1500;
  cr = #13;
type
  TCocoAction = ^TActionNode;
  TTarget = ^TTargetNode;
  TState = record (* state of finite automaton *)
    firstAction: TCocoAction; (* to first TCocoAction of this state *)
    endOf: Integer; (* nr. of recognized token if state is final *)
    ctx: boolean; (* TRUE: state reached by contextTrans *)
  end;
  TActionNode = record (* TCocoAction of finite automaton *)
    typ: TCocoNodeType; (* type of TCocoAction symbol: char, class *)
    sym: Integer; (* TCocoAction symbol *)
    tc: Integer; (* transition code: normTrans, contextTrans *)
    Target: TTarget; (* states after transition with input symbol *)
    Next: TCocoAction;
  end;
  TTargetNode = record (* state after transition with input symbol *)
    theState: Integer; (* .Target state *)
    Next: TTarget;
  end;


type
  TpbCocoAutomaton = class(TObject)
  private
    FdirtyDFA: boolean;
    FNewLine: boolean;
    FfirstComment: TComment;
    FrootState: Integer;
    FlastSimState: Integer;
    FlastState: Integer;
    FfirstMelted: TMelted;
    Fscanner: TStringStream;
    Ffram: TStringStream;
    FCocoScanner: TpbCocoScanner;
    FCocoTable: TpbCocoTable;
    stateArray: array[0..maxStates] of TState;
    FFrameInSize: Integer;
    FFramInPos: Integer;
    FStepOverWhiteSpace: boolean;
    function NewState: Integer;
    procedure SemErr(nr: Integer);
    procedure AddAction(act: TCocoAction; var head: TCocoAction);
    procedure AddTargetList(var lista, listb: TTarget);
    procedure ChangeAction(a: TCocoAction; sset: TCRTSet);
    procedure CombineShifts;
    procedure DeleteActionList(anAction: TCocoAction);
    procedure DeleteRedundantStates;
    procedure DeleteTargetList(list: TTarget);
    procedure DetachAction(a: TCocoAction; var L: TCocoAction);
    procedure GenComment(leftMarg: Integer; com: TComment);
    procedure GenLiterals(leftMarg: Integer);
    procedure Indent(n: Integer);
    procedure MakeSet(p: TCocoAction; var sset: TCRTSet);
    procedure MakeUnique(s: Integer; var changed: boolean);
    procedure MeltStates(s: Integer; var correct: boolean);
    procedure NewTransition(from: Integer; gn: TGraphNode;
      toState: Integer);
    procedure Put(ch: char);
    procedure PutB(n: Integer);
    procedure PutC(ch: char);
    procedure PutChCond(ch: char);
    procedure PutI(i: Integer);
    procedure PutI2(i, n: Integer);
    procedure PutLn;
    procedure PutRange(s: TCRTSet; indent: Integer);
    procedure PutS(s: string);
    procedure PutS1(s: string);
    procedure PutSE(i: Integer);
    procedure PutSN(i: Integer);
    procedure SetdirtyDFA(const Value: boolean);
    procedure SetlastSimState(const Value: Integer);
    procedure SetlastState(const Value: Integer);
    procedure SetNewLine(const Value: boolean);
    procedure SetrootState(const Value: Integer);
    procedure SplitActions(a, b: TCocoAction);
    procedure WriteState(leftMarg, s: Integer; var FirstState: boolean);
    function NewMelted(sset: TCRTSet; s: Integer): TMelted;
    procedure SetCocoScanner(const Value: TpbCocoScanner);
    function TheAction(theState: TState; ch: char): TCocoAction;
    procedure SetCocoTable(const Value: TpbCocoTable);
    procedure WriteLst(const S: string);
    procedure SetFrameInSize(const Value: Integer);
    function GetGenScannerStream: TStringStream;
    procedure SetFramInPos(const Value: Integer);
    procedure SetStepOverWhiteSpace(const Value: boolean);
    { Private declarations }
  protected
    { Protected declarations }
    //    SafeGuard : ISafeGuard;
    procedure CleanUpMemory;
  public
    { Public declarations }
    constructor Create;
    destructor Destroy; override;
    procedure CopyFramePart(stopStr: string; var leftMarg: Integer;
      framIn, framOut: TStringStream);
(* "stopStr" must not contain "FileIO.EOL".
   "leftMarg" is in/out-parameter  --  it has to be set once by the
   calling program.    *)
    procedure ImportSymConsts(leader: string; putS: TPutSProc);
    (* Generates the USES references for the eventually existing named constants. *)
    procedure ConvertToStates(gp0, sp: Integer);
(* Converts top-down graph with root gp into a subautomaton that
   recognizes token sp *)
    procedure MatchDFA(str: string; sp: Integer; var matchedSp: Integer);
(* Returns TRUE, if string str can be recognized by the current DFA.
   matchedSp is the token as that s can be recognized. *)
    procedure MakeDeterministic(var correct: boolean);
    (* Converts the NFA into a DFA. correct indicates if an error occurred. *)
    procedure NewComment(start, stop: Integer; nested: boolean);
(* Defines a new TComment for the fscanner. The TComment brackets are
   represented by the mini top-down graphs with the roots from and to. *)
    procedure WriteScanner(var ok: boolean);
(* Emits the source code of the generated fscanner using the frame file
   fscanner.frm *)

    procedure PrintStates;
    (* List the automaton for tracing *)
    procedure ClearMeltMemory;
    property FramInPos: Integer read FFramInPos write SetFramInPos;
  published
    { Published declarations }
    property lastSimState: Integer read FlastSimState write SetlastSimState;
    (* last non TMelted state *)
    property lastState: Integer read FlastState write SetlastState;
    (* last allocated state  *)
    property rootState: Integer read FrootState write SetrootState;
    (* start state of DFA    *)
    property dirtyDFA: boolean read FdirtyDFA write SetdirtyDFA;
    property NewLine: boolean read FNewLine write SetNewLine;
    property FrameInSize: Integer read FFrameInSize write SetFrameInSize;
    property CocoScanner: TpbCocoScanner read FCocoScanner write SetCocoScanner;
    property CocoTable: TpbCocoTable read FCocoTable write SetCocoTable;
    property GenScannerStream: TStringStream read GetGenScannerStream;
    property StepOverWhiteSpace: boolean read FStepOverWhiteSpace write SetStepOverWhiteSpace;
  end;

implementation

{ TpbCocoAutomaton }

constructor TpbCocoAutomaton.Create;
begin
  inherited;
  lastState := -1;
  rootState := NewState;
  FfirstMelted := nil;
  FfirstComment := nil;
  ffram := nil;
  NewLine := True;
  dirtyDFA := False;
  Fscanner := TStringStream.Create('');
  FStepOverWhiteSpace := True;
end;

destructor TpbCocoAutomaton.Destroy;
begin
  Fscanner.Free;
  Ffram.Free;
  CleanUpMemory;
  inherited;
end;

procedure TpbCocoAutomaton.SemErr(nr: Integer);
begin
  Cocoscanner.Error(nr + 100, Cocoscanner.line, Cocoscanner.col, Cocoscanner.pos)
end;

procedure TpbCocoAutomaton.Put(ch: char);
begin
  pbWrite(fscanner, ch);
end;

procedure TpbCocoAutomaton.PutLn;
begin
  pbWriteLn(fScanner, '');
end;

procedure TpbCocoAutomaton.PutB(n: Integer);
var
  S: string;
begin
  S := StringOfChar(' ', n);
  pbWriteLn(fScanner, S);
end;

procedure TpbCocoAutomaton.Indent(n: Integer);
begin
  if NewLine then PutB(n)
  else
    NewLine := True
end;

procedure TpbCocoAutomaton.PutS(s: string); far;
var
  i: integer;
begin
  for i := 1 to Length(s) do
    if s[i] = '$' then PutLn
  else
    Put(s[i]);
end;

procedure TpbCocoAutomaton.PutS1(s: string);
begin
  if s[1] = '"' then
  begin
    s[1] := '''';
    s[Length(s)] := ''''
  end;
  PutS(s);
end;

procedure TpbCocoAutomaton.PutI(i: Integer);
var
  S: string;
begin
  S := IntToStr(i);
  //  put(s[1]);
  putS(s);
end;

procedure TpbCocoAutomaton.PutI2(i, n: Integer);
var
  S: string;
begin
  S := IntToStr(i);
  pbWrite(fScanner, S + StringOfChar(' ', n));
end;

procedure TpbCocoAutomaton.PutC(ch: char);
begin
  case ch of
    #0..#31, #127..#255, '''':
    begin
      PutS('CHR(');
      PutI(Ord(ch));
      Put(')')
    end;
    else
    begin
      Put('''');
      Put(ch);
      Put('''')
    end
  end
end;

procedure TpbCocoAutomaton.PutSN(i: Integer);
var
  sn: TSymbolNode;
begin
  FCocoTable.GetSym(i, sn);
  if Length(sn.constant) > 0 then PutS(sn.constant)
  else
    PutI(i);
end;

procedure TpbCocoAutomaton.PutSE(i: Integer);
begin
  PutS('BEGIN sym := ');
  PutSN(i);
  PutS('; ');
end;

procedure TpbCocoAutomaton.PutRange(s: TCRTSet; indent: Integer);
var
  lo, hi: array[0..31] of char;
  top, i: integer;
  s1: TCRTSet;
begin
  (*----- fill lo and hi *)
  top := -1;
  i := 0;
  while i < 256 (*PDT*) do
  begin
    if IsIn(s, i) then
    begin
      INC(top);
      lo[top] := CHR(i);
      INC(i);
      while (i < 256 (*PDT*)) and IsIn(s, i) do INC(i);
      hi[top] := CHR(i - 1)
    end
    else
      INC(i)
  end;
  (*----- print ranges *)
  if (top = 1) and (lo[0] = #0) and (hi[1] = #255
    (*PDT*)) and (CHR(Ord(hi[0]) + 2) = lo[1]) then
  begin
    Fill(s1);
    Differ(s1, s);
    PutS('NOT ');
    PutRange(s1, indent);
  end
  else
  begin
    PutS('(');
    i := 0;
    while i <= top do
    begin
      if hi[i] = lo[i] then
      begin
        PutS('(ch = ');
        PutC(lo[i])
      end
      else if lo[i] = #0 then
      begin
        PutS('(ch <= ');
        PutC(hi[i])
      end
      else if hi[i] = #255 (*PDT*) then
      begin
        PutS('(ch >= ');
        PutC(lo[i])
      end
      else
      begin
        PutS('(ch >= ');
        PutC(lo[i]);
        PutS(') AND (ch <= ');
        PutC(hi[i])
      end;
      Put(')');
      if i < top then
      begin
        PutS(' OR$');
        PutB(indent)
      end;
      INC(i)
    end;
    Put(')')
  end
end;

procedure TpbCocoAutomaton.PutChCond(ch: char);
begin
  PutS('(ch = ');
  PutC(ch);
  Put(')')
end;

(* PrintStates          List the automaton for tracing
-------------------------------------------------------------------------*)

procedure TpbCocoAutomaton.PrintStates;

  procedure PrintSymbol(gnType: TCocoNodeType; val, Width: Integer);
  var
    sName: TsName;
    S: string;
    //    StrSize: integer;
    len: integer;
  begin
    { TODO :
              Changed from "if typ = FCocoTable.chClass then"
              to "if typ = Ord(FCocoTable.chClass) then".
              Needs testing. }
    if gnType = cnCharClass then
    begin
      FCocoTable.GetClassName(val, sName);
      WriteLst(sName);

      len := Length(sName)
    end
    else if (val >= Ord(' ')) and (val < 127) and (val <> 34) then
    begin
      WriteLst('"' + CHR(val) + '"');
      len := 3
    end
    else
    begin
      S := 'CHR(' + IntToStr(val) + ')';
      WriteLst(S);

      len := 7
    end;
    while len < Width do
    begin
      WriteLst(' ');
      INC(len)
    end
  end;
var
  anAction: TCocoAction;
  First: boolean;
  s, i: integer;
  targ: TTarget;
  sset: TCRTSet;
  Name: TsName;
begin
  WriteLst(#13#10);
  Cocoscanner.lst.Write('-------- states ---------', Length('-------- states ---------'));
  s := rootState;
  while s <= lastState do
  begin
    anAction := stateArray[s].firstAction;
    First := True;
    if stateArray[s].endOf = Ord(cnNoSym) then
      WriteLst('     ')
    else
      WriteLst('E(' + IntToStr(stateArray[s].endOf) + ')');

    if anAction = nil then
      pbWriteln(FCocoScanner.lst, '');
    while anAction <> nil do
    begin
      if First then
      begin
        WriteLst(' ');
        First := False
      end
      else
      begin
        WriteLst('          ');
      end;
      PrintSymbol(anAction^.typ, anAction^.sym, 0);
      WriteLst(' ');
      targ := anAction^.Target;
      while targ <> nil do
      begin
        WriteLst(IntToStr(targ^.theState) + ' ');
        targ := targ^.Next;
      end;
      if anAction^.tc = contextTrans then
        WriteLst(' context')
      else
        WriteLst(#13#10);
      anAction := anAction^.Next
    end;
    INC(s)
  end;
  WriteLst(#13#10);
  WriteLst('-------- character classes ---------');
  i := 0;
  while i <= FCocoTable.maxC do
  begin
    FCocoTable.GetClass(i, sset);
    FCocoTable.GetClassName(i, Name);
    WriteLst(Name +': ');
    Print(Cocoscanner.lst, sset, 80, 13);
    WriteLst(#13#10);
    pbWriteLn(Cocoscanner.lst, '');
    INC(i)
  end
end;

(* AddAction            Add a TCocoAction to the TCocoAction list of a state
------------------------------------------------------------------------*)

procedure TpbCocoAutomaton.AddAction(act: TCocoAction; var head: TCocoAction);
var
  a, lasta: TCocoAction;
begin
  a := head;
  lasta := nil;
  while True do
  begin
    if (a = nil) or (act^.typ < a^.typ) then
    (*collecting classes at the front improves performance*)
    begin
      act^.Next := a;
      if lasta = nil then head := act
      else
      begin
        lasta^.Next := act;
      end;
      EXIT;
    end;
    lasta := a;
    a := a^.Next;
  end;
end;

(* DetachAction         Detach TCocoAction a from list L
------------------------------------------------------------------------*)

procedure TpbCocoAutomaton.DetachAction(a: TCocoAction; var L: TCocoAction);
begin
  if L = a then L := a^.Next
  else if L <> nil then DetachAction(a, L^.Next)
end;

function TpbCocoAutomaton.TheAction(theState: TState; ch: char): TCocoAction;
var
  a: TCocoAction;
  sset: TCRTSet;
begin
  a := theState.firstAction;
  while a <> nil do
  begin
    if a^.typ = cnSingleChar then
    begin
      if Ord(ch) = a^.sym then
      begin
        TheAction := a;
        EXIT
      end
    end
    else if a^.typ = cnCharClass then
    begin
      FCocoTable.GetClass(a^.sym, sset);
      if IsIn(sset, Ord(ch)) then
      begin
        TheAction := a;
        EXIT
      end
    end;
    a := a^.Next
  end;
  TheAction := nil
end;

procedure TpbCocoAutomaton.AddTargetList(var lista, listb: TTarget);
var
  p, t: TTarget;


  procedure AddTarget(t: TTarget; var list: TTarget);
  label
    999;
  var
    p, lastp: TTarget;
  begin
    p := list;
    lastp := nil;
    while True do
    begin
      if (p = nil) or (t^.theState < p^.theState) then goto 999;
      if p^.theState = t^.theState then
      begin
        DISPOSE(t);
        EXIT
      end;
      lastp := p;
      p := p^.Next
    end;
    999:
      t^.Next := p;
    if lastp = nil then list := t
    else
      lastp^.Next := t
  end;
begin
  p := lista;
  while p <> nil do
  begin
    NEW(t);
    //t := GuardGetMem(SizeOf(TTarget),SafeGuard);
    //    Guard(t, SafeGuard);

    t^.theState := p^.theState;
    AddTarget(t, listb);
    p := p^.Next
  end
end;

(* NewMelted            Generate new info about a TMelted state
------------------------------------------------------------------------*)

function TpbCocoAutomaton.NewMelted(sset: TCRTSet; s: Integer): TMelted;
var
  melt: TMelted;
begin
  //  NEW(melt);
  //  Guard(melt,SafeGuard);
  NEW(melt);
  //  Guard(melt,SafeGuard);
  melt^.sset := sset;
  melt^.theState := s;
  melt^.Next := ffirstMelted;
  ffirstMelted := melt;
  NewMelted := melt
end;

(* NewState             Return a new state node
------------------------------------------------------------------------*)

function TpbCocoAutomaton.NewState: Integer;
begin
  INC(fLastState);
  if fLastState > maxStates then
    FCocoTable.Restriction(7, maxStates);
  stateArray[fLastState].firstAction := nil;
  stateArray[fLastState].endOf := Ord(cnNoSym);
  stateArray[fLastState].ctx := False;
  NewState := fLastState
end;

(* NewTransition        Generate transition (gn.theState, gn.p1) --> toState
------------------------------------------------------------------------*)

procedure TpbCocoAutomaton.NewTransition(from: Integer; gn: TGraphNode;
  toState: Integer);
var
  a: TCocoAction;
  t: TTarget;
begin
  if toState = rootState then SemErr(21);
  NEW(t);
  t^.theState := toState;
  t^.Next := nil;
  NEW(a);
  a^.typ := gn.gnType;
  a^.sym := gn.p1;
  a^.tc := gn.p2;
  a^.Target := t;
  AddAction(a, stateArray[from].firstAction);
end;

(* NewComment           Define new TComment
-------------------------------------------------------------------------*)

procedure TpbCocoAutomaton.NewComment(start, stop: Integer; nested: boolean);
var
  com: TComment;

  procedure MakeStr(gp: Integer; var s: STRING2);
  var
    i, n: integer;
    gn: TGraphNode;
    sset: TCRTSet;
  begin
    i := 1;
    while gp <> 0 do
    begin
      FCocoTable.GetNode(gp, gn);
      if gn.gnType = cnSingleChar then
      begin
        if i < 3 then s[i] := CHR(gn.p1);
        INC(i)
      end
      else if gn.gnType = cnCharClass then
      begin
        FCocoTable.GetClass(gn.p1, sset);
        if Elements(sset, n) <> 1 then SemErr(26);
        if i < 3 then s[i] := CHR(n);
        INC(i)
      end
      else
        SemErr(22);
      gp := gn.Next
    end;
    if (i = 1) or (i > 3) then SemErr(25)
    else
      s[0] := CHR(i - 1)
  end;
begin
  NEW(com);
  MakeStr(start, com^.start);
  MakeStr(stop, com^.stop);
  com^.nested := nested;
  com^.Next := ffirstComment;
  ffirstComment := com
end;

(* DeleteTargetList     Delete a .Target list
-------------------------------------------------------------------------*)

procedure TpbCocoAutomaton.DeleteTargetList(list: TTarget);
begin
  if list <> nil then
  begin
    DeleteTargetList(list^.Next);
    DISPOSE(list)
  end;
end;

(* DeleteActionList     Delete an TCocoAction list
-------------------------------------------------------------------------*)

procedure TpbCocoAutomaton.DeleteActionList(anAction: TCocoAction);
begin
  if anAction <> nil then
  begin
    DeleteActionList(anAction^.Next);
    DeleteTargetList(anAction^.Target);
    DISPOSE(anAction)
  end
end;

(* MakeSet              Expand TCocoAction symbol into symbol set
-------------------------------------------------------------------------*)

procedure TpbCocoAutomaton.MakeSet(p: TCocoAction; var sset: TCRTSet);
begin
  if p^.typ = cnCharClass then FCocoTable.GetClass(p^.sym, sset)
  else
  begin
    Clear(sset);
    Incl(sset, p^.sym)
  end
end;

(* ChangeAction         Change the TCocoAction symbol to set
-------------------------------------------------------------------------*)

procedure TpbCocoAutomaton.ChangeAction(a: TCocoAction; sset: TCRTSet);
var
  nr: integer;
begin
  if Elements(sset, nr) = 1 then
  begin
    a^.typ := cnSingleChar;
    a^.sym := nr
  end
  else
  begin
    nr := FCocoTable.ClassWithSet(sset);
    if nr < 0 then nr := FCocoTable.NewClass('##', sset);
    a^.typ := cnCharClass;
    a^.sym := nr
  end
end;

(* CombineShifts     Combine shifts with different symbols into same state
-------------------------------------------------------------------------*)

procedure TpbCocoAutomaton.CombineShifts;
var
  s: integer;
  a, b, c: TCocoAction;
  seta, setb: TCRTSet;
begin
  s := rootState;
  while s <= lastState do
  begin
    a := stateArray[s].firstAction;
    while a <> nil do
    begin
      b := a^.Next;
      while b <> nil do
      begin
        if (a^.Target^.theState = b^.Target^.theState) and (a^.tc = b^.tc) then
        begin
          MakeSet(a, seta);
          MakeSet(b, setb);
          Unite(seta, setb);
          ChangeAction(a, seta);
          c := b;
          b := b^.Next;
          DetachAction(c, a)
        end
        else
          b := b^.Next
      end;
      a := a^.Next
    end;
    INC(s)
  end
end;

(* DeleteRedundantStates   Delete unused and equal states
-------------------------------------------------------------------------*)

procedure TpbCocoAutomaton.DeleteRedundantStates;
var
  anAction: TCocoAction;
  s, s2, Next: integer;
  used: BITARRAY;
  {ARRAY [0 .. maxStates DIV size] OF BITSET }(*KJG*)
  newStateNr: array[0..maxStates] of integer;

  procedure FindUsedStates(s: Integer);
  var
    anAction: TCocoAction;
  begin
    if IsIn(used, s) then EXIT;
    Incl(used, s);
    anAction := stateArray[s].firstAction;
    while anAction <> nil do
    begin
      FindUsedStates(anAction^.Target^.theState);
      anAction := anAction^.Next
    end
  end;
begin
  Clear(used);
  FindUsedStates(rootState);
  (*---------- combine equal final states ------------*)
  s := rootState + 1;
  (*root state cannot be final*)
  while s <= lastState do
  begin
    if IsIn(used, s) and (stateArray[s].endOf <> Ord(cnNoSym)) then
      if (stateArray[s].firstAction = nil) and not stateArray[s].ctx then
      begin
        s2 := s + 1;
        while s2 <= lastState do
        begin
          if IsIn(used, s2) and (stateArray[s].endOf = stateArray[s2].endOf) then
            if (stateArray[s2].firstAction = nil) and not stateArray[s2].ctx then
            begin
              Excl(used, s2);
              newStateNr[s2] := s
            end;
          INC(s2)
        end
      end;
    INC(s)
  end;
  s := rootState;
    (* + 1 ?  PDT - was rootState, but Oberon had .next ie +1
                    seems to work both ways?? *)
  while s <= lastState do
  begin
    if IsIn(used, s) then
    begin
      anAction := stateArray[s].firstAction;
      while anAction <> nil do
      begin
        if not IsIn(used, anAction^.Target^.theState) then
          anAction^.Target^.theState := newStateNr[anAction^.Target^.theState];
        anAction := anAction^.Next
      end
    end;
    INC(s)
  end;
  (*-------- delete unused states --------*)
  s := rootState + 1;
  Next := s;
  while s <= lastState do
  begin
    if IsIn(used, s) then
    begin
      if Next < s then stateArray[Next] := stateArray[s];
      newStateNr[s] := Next;
      INC(Next)
    end
    else
      DeleteActionList(stateArray[s].firstAction);
    INC(s)
  end;
  lastState := Next - 1;
  s := rootState;
  while s <= lastState do
  begin
    anAction := stateArray[s].firstAction;
    while anAction <> nil do
    begin
      anAction^.Target^.theState := newStateNr[anAction^.Target^.theState];
      anAction := anAction^.Next
    end;
    INC(s)
  end
end;

(* ConvertToStates    Convert the TDG in gp into a subautomaton of the DFA
------------------------------------------------------------------------*)

procedure TpbCocoAutomaton.ConvertToStates(gp0, sp: Integer);
  (*note: gn.line is abused as a state number!*)
var
  stepped, visited: TMarkList;

  procedure NumberNodes(gp, snr: Integer);
  var
    gn: TGraphNode;
  begin
    if gp = 0 then EXIT; (*end of graph*)
    FCocoTable.GetNode(gp, gn);
    if gn.line >= 0 then EXIT; (*already visited*)
    if snr < rootState then snr := NewState;
    gn.line := snr;
    FCocoTable.PutNode(gp, gn);
    if FCocoTable.DelGraph(gp) then stateArray[snr].endOf := sp;
    (*snr is end state*)
    case gn.gnType of
      cnCharClass, cnSingleChar:
      begin
        NumberNodes(ABS(gn.Next), rootState - 1)
      end;
      cnOption:
      begin
        NumberNodes(ABS(gn.Next), rootState - 1);
        NumberNodes(gn.p1, snr)
      end;
      cnIteration:
      begin
        NumberNodes(ABS(gn.Next), snr);
        NumberNodes(gn.p1, snr)
      end;
      cnAlternative:
      begin
        NumberNodes(gn.p1, snr);
        NumberNodes(gn.p2, snr)
      end;
    end;
  end;

  function TheState(gp: Integer): Integer;
  var
    s: integer;
    gn: TGraphNode;
  begin
    if gp = 0 then
    begin
      s := NewState;
      stateArray[s].endOf := sp;
      TheState := s
    end
    else
    begin
      FCocoTable.GetNode(gp, gn);
      TheState := gn.line
    end
  end;

  procedure Step(from, gp: Integer);
  var
    gn: TGraphNode;
    Next: integer;
  begin
    if gp = 0 then EXIT;
    FCocoTable.InclMarkList(stepped, gp);
    FCocoTable.GetNode(gp, gn);
    case gn.gnType of
      cnCharClass, cnSingleChar:
      begin
        NewTransition(from, gn, TheState(ABS(gn.Next)))
      end;
      cnAlternative:
      begin
        Step(from, gn.p1);
        Step(from, gn.p2)
      end;
      cnOption, cnIteration:
      begin
        Next := ABS(gn.Next);
        if not FCocoTable.IsInMarkList(stepped, Next) then Step(from, Next);
        Step(from, gn.p1)
      end;
    end
  end;

  procedure FindTrans(gp: Integer; start: boolean);
  var
    gn: TGraphNode;
  begin
    if (gp = 0) or FCocoTable.IsInMarkList(visited, gp) then EXIT;
    FCocoTable.InclMarkList(visited, gp);
    FCocoTable.GetNode(gp, gn);
    if start then (* start of group of equally numbered nodes *)
    begin
      FCocoTable.ClearMarkList(stepped);
      Step(gn.line, gp);
    end;
    case gn.gnType of
      cnCharClass, cnSingleChar:
      begin
        FindTrans(ABS(gn.Next), True)
      end;
      cnOption:
      begin
        FindTrans(ABS(gn.Next), True);
        FindTrans(gn.p1, False)
      end;
      cnIteration:
      begin
        FindTrans(ABS(gn.Next), False);
        FindTrans(gn.p1, False)
      end;
      cnAlternative:
      begin
        FindTrans(gn.p1, False);
        FindTrans(gn.p2, False)
      end;
    end;
  end;
var
  gn: TGraphNode;
  i: integer;
begin
  if FCocoTable.DelGraph(gp0) then SemErr(20);
  for i := 0 to FCocoTable.nNodes do
  begin
    FCocoTable.GetNode(i, gn);
    gn.line := -1;
    FCocoTable.PutNode(i, gn)
  end;
  NumberNodes(gp0, rootState);
  SetLength(Visited, NodeSetSize);
  SetLength(Stepped, NodeSetSize);
  FCocoTable.ClearMarkList(visited);
  FindTrans(gp0, True)
end;


procedure TpbCocoAutomaton.MatchDFA(str: string; sp: Integer; var matchedSp: Integer);
label
  999;
var
  s, sto: integer (*State*);
  a: TCocoAction;
  gn: TGraphNode;
  i, len: integer;
  weakMatch: boolean;
begin (* s with quotes *)
  a := nil;
  s := rootState;
  i := 2;
  len := Length(str);
  weakMatch := False;
  while True do
  begin
    (* try to match str against existing DFA *)
    if i >= len then goto 999;
    if (S < maxstates) then
      a := TheAction(stateArray[s], str[i]);
    if a = nil then goto 999;
    if a^.typ = cnCharClass then weakMatch := True;
    s := a^.Target^.theState;
    INC(i)
  end;
  999:
    if weakMatch and (i < len) then
    begin
      s := rootState;
      i := 2;
      dirtyDFA := True
    end;
  while i < len do
  begin
    (* make new DFA for str[i..len-1] *)
    sto := NewState;
    gn.gnType := cnSingleChar;
    gn.p1 := Ord(str[i]);
    gn.p2 := normTrans;
    NewTransition(s, gn, sto);
    s := sto;
    INC(i)
  end;
  matchedSp := stateArray[s].endOf;
  if stateArray[s].endOf = Ord(cnNoSym) then stateArray[s].endOf := sp;
end;

(* SplitActions     Generate unique actions from two overlapping actions
-----------------------------------------------------------------------*)

procedure TpbCocoAutomaton.SplitActions(a, b: TCocoAction);
var
  c: TCocoAction;
  seta, setb, setc: TCRTSet;

  procedure CombineTransCodes(t1, t2: Integer; var Result: Integer);
  begin
    if t1 = contextTrans then Result := t1
    else
      Result := t2
  end;
begin
  MakeSet(a, seta);
  MakeSet(b, setb);
  if Equal(seta, setb) then
  begin
    AddTargetList(b^.Target, a^.Target);
    DeleteTargetList(b^.Target);
    CombineTransCodes(a^.tc, b^.tc, a^.tc);
    DetachAction(b, a);
    DISPOSE(b);
  end
  else if Includes(seta, setb) then
  begin
    setc := seta;
    Differ(setc, setb);
    AddTargetList(a^.Target, b^.Target);
    CombineTransCodes(a^.tc, b^.tc, b^.tc);
    ChangeAction(a, setc)
  end
  else if Includes(setb, seta) then
  begin
    setc := setb;
    Differ(setc, seta);
    AddTargetList(b^.Target, a^.Target);
    CombineTransCodes(a^.tc, b^.tc, a^.tc);
    ChangeAction(b, setc)
  end
  else
  begin
    Intersect(seta, setb, setc);
    Differ(seta, setc);
    Differ(setb, setc);
    ChangeAction(a, seta);
    ChangeAction(b, setb);
    NEW(c);
    c^.Target := nil;
    CombineTransCodes(a^.tc, b^.tc, c^.tc);
    AddTargetList(a^.Target, c^.Target);
    AddTargetList(b^.Target, c^.Target);
    ChangeAction(c, setc);
    AddAction(c, a)
  end
end;

(* MakeUnique           Make all actions in this state unique
-------------------------------------------------------------------------*)

procedure TpbCocoAutomaton.MakeUnique(s: Integer; var changed: boolean);
var
  a, b: TCocoAction;

  function Overlap(a, b: TCocoAction): boolean;
  var
    seta, setb: TCRTSet;
  begin
    if a^.typ = cnSingleChar then
    begin
      if b^.typ = cnSingleChar then
      begin
        Overlap := a^.sym = b^.sym
      end
      else
      begin
        FCocoTable.GetClass(b^.sym, setb);
        Overlap := IsIn(setb, a^.sym)
      end
    end
    else
    begin
      FCocoTable.GetClass(a^.sym, seta);
      if b^.typ = cnSingleChar then
      begin
        Overlap := IsIn(seta, b^.sym)
      end
      else
      begin
        FCocoTable.GetClass(b^.sym, setb);
        Overlap := not Different(seta, setb)
      end
    end
  end;
begin
  a := stateArray[s].firstAction;
  changed := False;
  while a <> nil do
  begin
    b := a^.Next;
    while b <> nil do
    begin
      if Overlap(a, b) then
      begin
        SplitActions(a, b);
        changed := True;
        EXIT
              (* originally no RETURN.  FST blows up if we leave RETURN out.
                 Somewhere there is a field that is not properly set, but I
                 have not chased this down completely Fri  08-20-1993 *)
      end;
      b := b^.Next;
    end;
    a := a^.Next
  end;
end;

(* MeltStates       Melt states appearing with a shift of the same symbol
-----------------------------------------------------------------------*)

procedure TpbCocoAutomaton.MeltStates(s: Integer; var correct: boolean);
var
  anAction: TCocoAction;
  ctx: boolean;
  endOf: integer;
  melt: TMelted;
  sset: TCRTSet;
  s1: integer;
  changed: boolean;


  procedure AddMeltedSet(nr: Integer; var sset: TCRTSet);
  var
    m: TMelted;
  begin
    m := ffirstMelted;
    while (m <> nil) and (m^.theState <> nr) do m := m^.Next;
    if m = nil then FCocoTable.Restriction(-1, 0);
    Unite(sset, m^.sset);
  end;

  procedure GetStateSet(t: TTarget; var sset: TCRTSet;
    var endOf: Integer;
    var ctx: boolean);
  begin
    Clear(sset);
    endOf := Ord(cnNoSym);
    ctx := False;
    while t <> nil do
    begin
      if t^.theState <= lastSimState then Incl(sset, t^.theState)
      else
        AddMeltedSet(t^.theState, sset);
      if stateArray[t^.theState].endOf <> Ord(cnNoSym) then
      begin
        if (endOf = Ord(cnNoSym)) or (endOf = stateArray[t^.theState].endOf) then
        begin
          endOf := stateArray[t^.theState].endOf;
        end
        else
        begin
          WriteLst(#13#10);
          WriteLst('Tokens ' + IntToStr(endOf) + ' and ' +
            IntToStr(stateArray[t^.theState].endOf) + ' cannot be distinguished.');
          correct := False;
        end;
      end;
      if stateArray[t^.theState].ctx then
      begin
        ctx := True;
        if stateArray[t^.theState].endOf <> Ord(cnNoSym) then
        begin
          WriteLst(#13#10);
          WriteLst('Ambiguous CONTEXT clause.');
          correct := False
        end
      end;
      t := t^.Next
    end
  end;

  procedure FillWithActions(s: Integer; targ: TTarget);
  var
    anAction, a: TCocoAction;
  begin
    while targ <> nil do
    begin
      anAction := stateArray[targ^.theState].firstAction;
      while anAction <> nil do
      begin
        New(A);
        a^ := anAction^;
        a^.Target := nil;
        AddTargetList(anAction^.Target, a^.Target);
        AddAction(a, stateArray[s].firstAction);
        anAction := anAction^.Next
      end;
      targ := targ^.Next
    end;
  end;

  function KnownMelted(sset: TCRTSet; var melt: TMelted): boolean;
  begin
    melt := ffirstMelted;
    while melt <> nil do
    begin
      if Equal(sset, melt^.sset) then
      begin
        KnownMelted := True;
        EXIT
      end;
      melt := melt^.Next
    end;
    KnownMelted := False
  end;
begin
  anAction := stateArray[s].firstAction;
  while anAction <> nil do
  begin
    if anAction^.Target^.Next <> nil then
    begin
      GetStateSet(anAction^.Target, sset, endOf, ctx);
      if not KnownMelted(sset, melt) then
      begin
        s1 := NewState;
        stateArray[s1].endOf := endOf;
        stateArray[s1].ctx := ctx;
        FillWithActions(s1, anAction^.Target);
        repeat
          MakeUnique(s1, changed)
        until not changed;
        melt := NewMelted(sset, s1);
      end;
      DeleteTargetList(anAction^.Target^.Next);
      anAction^.Target^.Next := nil;
      anAction^.Target^.theState := melt^.theState
    end;
    anAction := anAction^.Next;
  end;
  //  ClearMemory;
end;

(* MakeDeterministic     Make NDFA --> DFA
------------------------------------------------------------------------*)

procedure TpbCocoAutomaton.MakeDeterministic(var correct: boolean);
var
  s: integer;
  changed: boolean;

  procedure FindCtxStates;
    (* Find states reached by a context transition *)
  var
    a: TCocoAction;
    s: integer;
  begin
    s := rootState;
    while s <= lastState do
    begin
      a := stateArray[s].firstAction;
      while a <> nil do
      begin
        if a^.tc = contextTrans then
          stateArray[a^.Target^.theState].ctx := True;
        a := a^.Next
      end;
      INC(s)
    end;
  end;
begin
  lastSimState := lastState;
  FindCtxStates;
  s := rootState;
  while s <= lastState do
  begin
    repeat
      MakeUnique(s, changed)
    until not changed;
    INC(s)
  end;
  correct := True;
  s := rootState;
  while s <= lastState do
  begin
    MeltStates(s, correct);
    INC(s)
  end;
  DeleteRedundantStates;
  CombineShifts;
  (* ====    IF FCocoTable.ddt["A"] THEN PrintStates END ==== *)
end;

(* GenComment            Generate a PROCEDURE TpbCocoAutomaton.to scan comments
-------------------------------------------------------------------------*)

procedure TpbCocoAutomaton.GenComment(leftMarg: Integer; com: TComment);

  procedure GenBody(leftMarg: Integer);
  begin
    PutB(leftMarg);
    PutS('WHILE TRUE DO BEGIN$');
    PutB(leftMarg + 2);
    PutS('IF ');
    PutChCond(com^.stop[1]);
    PutS(' THEN BEGIN$');
    if Length(com^.stop) = 1 then
    begin
      PutB(leftMarg + 4);
      PutS('DEC(level); oldEols := curLine - startLine; NextCh;$');
      PutB(leftMarg + 4);
      PutS('IF level = 0 THEN BEGIN Comment := TRUE; GOTO 999; END;$');
    end
    else
    begin
      PutB(leftMarg + 4);
      PutS('NextCh;$');
      PutB(leftMarg + 4);
      PutS('IF ');
      PutChCond(com^.stop[2]);
      PutS(' THEN BEGIN$');
      PutB(leftMarg + 6);
      PutS('DEC(level); NextCh;$');
      PutB(leftMarg + 6);
      PutS('IF level = 0 THEN BEGIN Comment := TRUE; GOTO 999; END$');
      PutB(leftMarg + 4);
      PutS('END$');
    end;
    if com^.nested then
    begin
      PutB(leftMarg + 2);
      PutS('END ELSE IF ');
      PutChCond(com^.start[1]);
      PutS(' THEN BEGIN$');
      if Length(com^.start) = 1 then
      begin
        PutB(leftMarg + 4);
        PutS('INC(level); NextCh;$');
      end
      else
      begin
        PutB(leftMarg + 4);
        PutS('NextCh;$');
        PutB(leftMarg + 4);
        PutS('IF ');
        PutChCond(com^.start[2]);
        PutS(' THEN BEGIN ');
        PutS('INC(level); NextCh ');
        PutS('END$');
      end;
    end;
    PutB(leftMarg + 2);
    PutS('END ELSE IF ch = EF THEN BEGIN Comment := FALSE; GOTO 999; END$');
    PutB(leftMarg + 2);
    PutS('ELSE NextCh;$');
    PutB(leftMarg);
    PutS('END; (* WHILE TRUE *)$');
  end;
begin
  PutS('IF ');
  PutChCond(com^.start[1]);
  PutS(' THEN BEGIN$');
  if Length(com^.start) = 1 then
  begin
    PutB(leftMarg + 2);
    PutS('NextCh;$');
    GenBody(leftMarg + 2)
  end
  else
  begin
    PutB(leftMarg + 2);
    PutS('NextCh;$');
    PutB(leftMarg + 2);
    PutS('IF ');
    PutChCond(com^.start[2]);
    PutS(' THEN BEGIN$');
    PutB(leftMarg + 4);
    PutS('NextCh;$');
    GenBody(leftMarg + 4);
    PutB(leftMarg + 2);
    PutS('END ELSE BEGIN$');
    PutB(leftMarg + 4);
    PutS('IF (ch = _CR) OR (ch = _LF) THEN BEGIN$');
    PutB(leftMarg + 6);
    PutS('DEC(curLine); lineStart := oldLineStart$');
    PutB(leftMarg + 4);
    PutS('END;$');
    PutB(leftMarg + 4);
    PutS('DEC(bp); ch := lastCh;$');
    PutB(leftMarg + 2);
    PutS('END;$');
  end;
  PutB(leftMarg);
  PutS('END;$');
  PutB(leftMarg);
end;

(* CopyFramePart   Copy from file <ffram> to file <framOut> until <stopStr>
-------------------------------------------------------------------------*)

procedure TpbCocoAutomaton.CopyFramePart(stopStr: string; var leftMarg: Integer;
  framIn, framOut: TStringStream);
const
  CR = #13;
  LF = #10;
var
  ch, startCh: char;
  slen, i, j: integer;
  temp: array[1..63] of char;
begin
  startCh := stopStr[1];
  pbRead(FramIn, ch);
  slen := Length(stopStr);
  inc(FFraminPos);
  while FFramInPos <= FFrameInSize do
  begin
    if (ch = CR) or (ch = LF) then leftMarg := 0
    else
      INC(leftMarg);
    if ch = startCh then (* check if stopString occurs *)
    begin
      i := 1;
      while (i < slen) and (ch = stopStr[i]) and (FFramInPos <= FFrameInSize) do
      begin
        temp[i] := ch;
        INC(i);
        pbRead(FramIn, ch);
        inc(FFraminPos);
      end;
      if ch = stopStr[i] then
      begin
        DEC(leftMarg);
        EXIT
      end;
      (* found ==> exit , else continue *)
      for j := 1 to i - 1 do
        pbWrite(framOut, temp[j]);
      pbWrite(Framout, ch);
      INC(leftMarg, i);
    end
    else
      pbWrite(Framout, ch);
    pbRead(FramIn, ch);
    inc(FFramInPos);
  end;
end;

(* ImportSymConsts      Generates the import of the named symbol constants
-------------------------------------------------------------------------*)

procedure TpbCocoAutomaton.ImportSymConsts(leader: string; putS: TPutSProc);
var
  oldLen, pos: integer;
  cname: TsName;
  gn: TGraphNode;
  sn: TSymbolNode;
  gramName: string;

  procedure PutImportSym;
  begin
    if pos + oldLen > MaxSourceLineLength then
    begin
      putS('$  ');
      pos := 2
    end;
    putS(cname);
    INC(pos, oldLen + 1);
      (* This is not strictly correct, as the increase of 2 should be
         lower. I omitted it, because to separate it would be too
         complicated, and no unexpected side effects are likely, since it
         is only called again outside the loop - after which "pos" is not
         used again
      *)
  end;
begin
  (* ----- Import list of the generated Symbol Constants Module ----- *)
  FCocoTable.GetNode(FCocoTable.root, gn);
  FCocoTable.GetSym(gn.p1, sn);
  putS(leader);
  gramName := sn.Name;
  putS(gramName);
  putS('G (* Symbol Constants *);$');
end;

(* GenLiterals           Generate CASE for the recognition of literals
-------------------------------------------------------------------------*)

procedure TpbCocoAutomaton.GenLiterals(leftMarg: Integer);
var
  i, j, k: integer;
  key: array[0..maxLiterals] of TsName;
  knr: array[0..maxLiterals] of integer;
  ch: char;
  sn: TSymbolNode;
begin
  (*-- sort literal list*)
  i := 0;
  k := 0;
  while i <= FCocoTable.maxT do
  begin
    FCocoTable.GetSym(i, sn);
    if sn.struct = litToken then
    begin
      j := k - 1;
      while (j >= 0) and (sn.Name < key[j]) do
      begin
        key[j + 1] := key[j];
        knr[j + 1] := knr[j];
        DEC(j)
      end;
      key[j + 1] := sn.Name;
      knr[j + 1] := i;
      INC(k);
      if k > maxLiterals then
        FCocoTable.Restriction(10, maxLiterals);
    end;
    INC(i)
  end;
  (*-- print CASE statement*)
  if k <> 0 then
  begin
    PutS('CurrentCh(bp0,cc);$');
    PutS('CASE cc OF$');
    PutB(leftMarg);
    i := 0;
    while i < k do
    begin
      ch := key[i, 2]; (*key[i, 0] = quote*)
      if i <> 0 then
      begin
        PutLn;
        PutB(leftMarg)
      end;
      PutS('  ');
      PutC(ch);
      j := i;
      repeat
        if i = j then PutS(': IF')
        else
        begin
          PutB(leftMarg + 6);
          PutS(' END ELSE IF')
        end;
        PutS(' Equal(');
        PutS1(key[i]);
        PutS(') THEN ');
        PutSE(knr[i]);
        PutLn;
        INC(i);
      until (i = k) or (key[i, 2] <> ch);
      PutB(leftMarg + 6);
      PutS(' END;');
    end;
    PutLn;
    PutB(leftMarg);
    PutS('ELSE BEGIN END$');
    PutB(leftMarg);
    PutS('END')
  end;
end;

(* WriteState           Write the source text of a fscanner state
-------------------------------------------------------------------------*)

procedure TpbCocoAutomaton.WriteState(leftMarg, s: Integer; var FirstState: boolean);
var
  anAction: TCocoAction;
  ind: integer;
  First, ctxEnd: boolean;
  sn: TSymbolNode;
  endOf: integer;
  sset: TCRTSet;
begin
  endOf := stateArray[s].endOf;
  if (endOf > FCocoTable.maxT) and (endOf <> Ord(cnNoSym)) then
  (*pragmas have been moved*)
  begin
    endOf := FCocoTable.maxT + maxSymbols - endOf
  end;
  Indent(leftMarg);
  if FirstState then FirstState := False;
  PutS('  ');
  PutI2(s, 2);
  PutS(': ');
  First := True;
  ctxEnd := stateArray[s].ctx;
  anAction := stateArray[s].firstAction;
  while anAction <> nil do
  begin
    if First then
    begin
      PutS('IF ');
      First := False;
      ind := leftMarg + 3
    end
    else
    begin
      PutB(leftMarg + 6);
      PutS('END ELSE IF ');
      ind := leftMarg + 6
    end;
    if anAction^.typ = cnSingleChar then
    begin
      PutChCond(CHR(anAction^.sym))
    end
    else
    begin
      FCocoTable.GetClass(anAction^.sym, sset);
      PutRange(sset, leftMarg + ind)
    end;
    PutS(' THEN BEGIN');
    if anAction^.Target^.theState <> s then
    begin
      PutS(' state := ');
      PutI(anAction^.Target^.theState);
      Put(';')
    end;
    if anAction^.tc = contextTrans then
    begin
      PutS(' INC(apx)');
      ctxEnd := False
    end
    else if stateArray[s].ctx then PutS(' apx := 0');
    PutS(' $');
    anAction := anAction^.Next
  end;
  if stateArray[s].firstAction <> nil then
  begin
    PutB(leftMarg + 6);
    PutS('END ELSE ')
  end;
  if endOf = Ord(cnNoSym) then
  begin
    PutS('BEGIN sym := no_Sym; ');
  end
  else (*final theState*)
  begin
    FCocoTable.GetSym(endOf, sn);
    if ctxEnd then (*cut appendix*)
    begin
      PutS('BEGIN bp := bp - apx - 1;');
      PutS(' DEC(fnextLen, apx); NextCh; ')
    end;
    PutSE(endOf);
    if sn.struct = classLitToken then
    begin
      PutS('CheckLiteral; ')
    end
  end;
  if ctxEnd then
  begin
    PutS('EXIT; END; END;$')
  end
  else
  begin
    PutS('EXIT; END;$');
  end;
end;

(* WriteScanner         Write the fscanner source file
-------------------------------------------------------------------------*)

procedure TpbCocoAutomaton.WriteScanner(var ok: boolean);
const
  ListingWidth = 78;
var
  gramName, fGramName, fn: string;
  startTab: array[0..255] of integer;
  com: TComment;
  i, j, s: integer;
  gn: TGraphNode;
  sn: TSymbolNode;

  procedure FillStartTab;
  var
    anAction: TCocoAction;
    i, targetState, undefState: integer;
    _class: TCRTSet;
  begin
    undefState := lastState + 2;
    startTab[0] := lastState + 1; (*eof*)
    i := 1;
    while i < 256 (*PDT*) do
    begin
      startTab[i] := undefState;
      INC(i)
    end;
    anAction := stateArray[rootState].firstAction;
    while anAction <> nil do
    begin
      targetState := anAction^.Target^.theState;
      if anAction^.typ = cnSingleChar then startTab[anAction^.sym] := targetState
      else
      begin
        FCocoTable.GetClass(anAction^.sym, _class);
        i := 0;
        while i < 256 (*PDT*) do
        begin
          if IsIn(_class, i) then startTab[i] := targetState;
          INC(i)
        end
      end;
      anAction := anAction^.Next
    end
  end;
var
  LeftMargin: integer;
  FirstState: boolean;
  ScannerFrame: string;
begin
  FStepOverWhiteSpace := not FCocoTable.ddt['W'];
  if dirtyDFA then
  begin
    MakeDeterministic(ok);
  end;
  FillStartTab;
  if (not CocoTable.hasScannerFrame) then
    ScannerFrame := Concat(Cocoscanner.directory, '\pbscanner.frm')
  else 
    ScannerFrame := CocoTable.ScannerFrame;
  try
    ffram := LoadStringStream(ScannerFrame);
    FFrameInSize := Ffram.Size;
  except
    raise Exception.Create('"' + ScannerFrame + '" not found - aborted.');
  end;
  LeftMargin := 0;
  FCocoTable.GetNode(FCocoTable.root, gn);
  FCocoTable.GetSym(gn.p1, sn);
  gramName := sn.Name;
  if (not CocoTable.hasScannerTarget) then
    fGramName := Concat(Cocoscanner.outputDirectory, '\' + gramName)
  else
    fGramName := CocoTable.ScannerTarget + '\' + gramName;


  (*------- *s.pas -------*)
  // for the unit title
  PutS('// Generated by ParserBuilder @ ' + DateTimeToStr(NOW) + '$$');
  CopyFramePart('-->modulename', LeftMargin, ffram, fscanner);
  PutS(gramName);

  // for the scanner's class name
  CopyFramePart('-->modulename', LeftMargin, ffram, fscanner);
  PutS(gramName);

  CopyFramePart('-->unitname', LeftMargin, ffram, fscanner);


  if FCocoTable.ddt['N'] or FCocoTable.symNames then ImportSymConsts('USES ', PutS);
  CopyFramePart('-->unknownsym', LeftMargin, ffram, fscanner);
  if FCocoTable.ddt['N'] or FCocoTable.symNames then PutSN(FCocoTable.maxT)
  else
    PutI(FCocoTable.maxT);

  // For CapChAt Method
  CopyFramePart('-->modulename', LeftMargin, ffram, fscanner);
  PutS(gramName);
  // For CharAt Method
  CopyFramePart('-->modulename', LeftMargin, ffram, fscanner);
  PutS(gramName);

  // For Comment Method

  CopyFramePart('-->modulename', LeftMargin, ffram, fscanner);
  PutS(gramName);



  CopyFramePart('-->comment', LeftMargin, ffram, fscanner);
  com := ffirstComment;
  while com <> nil do
  begin
    GenComment(LeftMargin, com);
    com := com^.Next;
  end;

  // For Err Method
  CopyFramePart('-->modulename', LeftMargin, ffram, fscanner);
  PutS(gramName);

  // For Get Method
  CopyFramePart('-->modulename', LeftMargin, ffram, fscanner);
  PutS(gramName);

  CopyFramePart('-->literals', LeftMargin, ffram, fscanner);
  GenLiterals(LeftMargin);

  CopyFramePart('-->GetSy1', LeftMargin, ffram, fscanner);
  NewLine := False;
  if not IsIn(FCocoTable.ignored, Ord(cr)) then
  begin
    Indent(LeftMargin);
    PutS('IF oldEols > 0 THEN BEGIN DEC(bp);');
    PutS(' DEC(oldEols); ch := _CR END;$')
  end;
  Indent(LeftMargin);
  if (FStepOverWhiteSpace) then
    PutS('WHILE (ch = '' '')')
  else
    PutS('WHILE $');

  if not Empty(FCocoTable.ignored) then
  begin
    if (FStepOverWhiteSpace) then
      PutS(' OR$');
    Indent(LeftMargin + 6);
    PutRange(FCocoTable.ignored, LeftMargin + 6);
  end;
  PutS(' DO NextCh;');
  if ffirstComment <> nil then
  begin
    PutLn;
    PutB(LeftMargin);
    PutS('IF (');
    com := ffirstComment;
    while com <> nil do
    begin
      PutChCond(com^.start[1]);
      if com^.Next <> nil then PutS(' OR ');
      com := com^.Next
    end;
    PutS(') AND Comment THEN BEGIN Get(sym); EXIT; END;');
  end;
  CopyFramePart('-->GetSy2', LeftMargin, ffram, fscanner);
  NewLine := False;
  s := rootState + 1;
  FirstState := True;
  while s <= lastState do
  begin
    WriteState(LeftMargin, s, FirstState);
    INC(s)
  end;
  PutB(LeftMargin);
  PutS('  ');
  PutI2(lastState + 1, 2);
  PutS(': ');
  PutSE(0);
  PutS('ch := #0; DEC(bp); EXIT END;');

  // for the GetName Method
  CopyFramePart('-->modulename', LeftMargin, ffram, fscanner);
  PutS(gramName);

  // for the GetString Method
  CopyFramePart('-->modulename', LeftMargin, ffram, fscanner);
  PutS(gramName);

  // for the NextCh Method
  CopyFramePart('-->modulename', LeftMargin, ffram, fscanner);
  PutS(gramName);


  // for the _Reset Method
  CopyFramePart('-->modulename', LeftMargin, ffram, fscanner);
  PutS(gramName);

  // for the Create Method
  CopyFramePart('-->modulename', LeftMargin, ffram, fscanner);
  PutS(gramName);

  CopyFramePart('-->initializations', LeftMargin, ffram, fscanner);
  if FCocoTable.ignoreCase then PutS('CurrentCh := CapChAt;$')
  else
    PutS('CurrentCh := CharAt;$');
  PutB(LeftMargin);
  i := 0;
  while i < 64 (*PDT*) do
  begin
    if i <> 0 then
    begin
      PutLn;
      PutB(LeftMargin);
    end;
    j := 0;
    while j < 4 do
    begin
      PutS('start[');
      PutI2(4 * i + j, 3);
      PutS('] := ');
      PutI2(startTab[4 * i + j], 2);
      PutS('; ');
      INC(j);
    end;
    INC(i);
  end;

  // for the Destroy Method
  CopyFramePart('-->modulename', LeftMargin, ffram, fscanner);
  PutS(gramName);

  // for the DisposeBuf Method
  CopyFramePart('-->modulename', LeftMargin, ffram, fscanner);
  PutS(gramName);

  // final comment to name the file at the bottom
  CopyFramePart('-->modulename', LeftMargin, ffram, fscanner);
  PutS(gramName + 'Scanner *)$');
  fn := Concat(fGramName, 'Scanner.pas');
  if FileExists(fn) then
    CopyFile(PChar(fn), PChar(ChangeFileExt(fn, '.pas.backup')), False);

  SaveStringStream(fScanner, fn);
end;

procedure TpbCocoAutomaton.SetdirtyDFA(const Value: boolean);
begin
  FdirtyDFA := Value;
end;

procedure TpbCocoAutomaton.SetlastSimState(const Value: Integer);
begin
  FlastSimState := Value;
end;

procedure TpbCocoAutomaton.SetlastState(const Value: Integer);
begin
  FlastState := Value;
end;

procedure TpbCocoAutomaton.SetNewLine(const Value: boolean);
begin
  FNewLine := Value;
end;

procedure TpbCocoAutomaton.SetrootState(const Value: Integer);
begin
  FrootState := Value;
end;

procedure TpbCocoAutomaton.SetCocoScanner(const Value: TpbCocoScanner);
begin
  FCocoScanner := Value;
end;

procedure TpbCocoAutomaton.SetCocoTable(const Value: TpbCocoTable);
begin
  FCocoTable := Value;
end;


procedure TpbCocoAutomaton.WriteLst(const S: string);
begin
  pbWrite(Cocoscanner.lst, S);
end;

procedure TpbCocoAutomaton.SetFrameInSize(const Value: Integer);
begin
  FFrameInSize := Value;
end;

function TpbCocoAutomaton.GetGenScannerStream: TStringStream;
begin
  Result := Fscanner;
end;

procedure TpbCocoAutomaton.SetFramInPos(const Value: Integer);
begin
  FFramInPos := Value;
end;

procedure TpbCocoAutomaton.SetStepOverWhiteSpace(const Value: boolean);
begin
  FStepOverWhiteSpace := Value;
end;

procedure TpbCocoAutomaton.CleanUpMemory;
var
  anAction: TCocoAction;
  s: integer;
begin
  s := rootState;
  while s <= lastState do
  begin
    anAction := stateArray[s].firstAction;
    if anAction <> nil then
    begin
      DeleteActionList(anAction);
    end;
    INC(s)
  end;
  ClearMeltMemory;
end;

procedure TpbCocoAutomaton.ClearMeltMemory;
var 
  melt, NextMelt: TMelted;
begin
  melt := ffirstMelted;
  while melt <> nil do
  begin
    NextMelt := Melt;
    melt := melt^.Next;
    Dispose(NextMelt);
  end;
end;


end.
