{-----------------------------------------------------------------------------
 Unit Name: pbCocoParserGen
 Author:     Originally Pat Terry (CRX.PAS), ported to Delphi by Pat Connors,
   Parser Generation
   =================

   Uses the top-down graph and the computed sets of terminal start symbols
   from pbCocoTable to generate a recursive descent parsing class.

   Errors are reported by error numbers. The corresponding error messages
   are written to <grammar name>.err.



-----------------------------------------------------------------------------}


unit pbCocoParserGen;

interface

uses
  Windows, Messages, SysUtils, Classes,
  pbCocoScanner, pbStrLib, pbCocoParser, pbCocoTable,
  pbCocoSets, pbCocoAutomaton, pbCocoConstants;


type
  TIndentProcType = procedure(i: Integer) of object;

type
  TpbCocoParserGen = class(TObject)
  private
    FCocoAutomaton: TpbCocoAutomaton;
    FCocoScanner: TpbCocoScanner;
    FCocoTable: TpbCocoTable;
    framinSize: Integer;
    FErrorFormat: string;
    function Alternatives(gp: Integer): Integer;
    function NewCondSet(newSet: TCRTSet): Integer;
    procedure SetCocoAutomaton(const Value: TpbCocoAutomaton);
    procedure SetCocoScanner(const Value: TpbCocoScanner);
    procedure SetCocoTable(const Value: TpbCocoTable);
    procedure SetErrorFormat(const Value: string);
    { Private declarations }
  protected
    { Protected declarations }
  public
    { Public declarations }
    symSet: array[0..symSetSize] of TCRTSet;
    (* symbol sets in the   generated parser *)
    maxSS: Integer; (* number of symbol sets *)
    errorNr: Integer; (* number of last generated error message*)
    curSy: Integer; (* symbol whose production is currently generated *)
    err: TStringStream; (* output: error message texts *)
    fram: TStringStream; (* input:  parser frame parser.frm *)
    syn: TStringStream; (* output: generated parser *)
    NewLine: boolean;
    IndDisp: Integer;
    ThegramName: string;
    constructor Create;
    destructor Destroy; override;
    procedure GenCompiler;
    procedure WriteStatistics;
    procedure GenPragmaCode(leftMarg: Integer; gramName: string);
    procedure InitSets;
    procedure GenProductions;
    procedure GenForwardRefs;
    procedure GenProcedureHeading(sn: TSymbolNode);
    procedure GenClassProcedureHeading(sn: TSymbolNode);
    procedure GenCode(gp, indent: Integer; Checked: TCRTSet);
    procedure GenCond(newSet: TCRTSet; indent: Integer);
    procedure GenErrorMsg(errTyp, errSym: Integer; var errNr: Integer);
    procedure CopySourcePart(pos: TPosition;
      indent: Integer; indentProc: TIndentProcType);
    procedure CopyFramePart(stopStr: string; var leftMarg: Integer);
    procedure PutSet1(s: TCRTSet);
    procedure PutSet(s: BITSET; offset: Integer);
    procedure PutSI(i: Integer);
    procedure PutI2(i: Integer);
    procedure PutI(i: Integer);
    procedure PutS(s: string);
    procedure IndentProc(n: Integer);
    procedure Indent(n: Integer);
    procedure PutB(n: Integer);
    procedure PutLn;
    procedure Put(ch: char);
  published
    property CocoScanner: TpbCocoScanner read FCocoScanner write SetCocoScanner;
    property CocoTable: TpbCocoTable read FCocoTable write SetCocoTable;
    property CocoAutomaton: TpbCocoAutomaton read FCocoAutomaton write SetCocoAutomaton;
    property ErrorFormat: string read FErrorFormat write SetErrorFormat;
  end;


implementation




{ TpbCocoParserGen }

constructor TpbCocoParserGen.Create;
begin
  inherited;
  errorNr := -1;
  maxSS := 0; (*symSet[0] reserved for allSyncSyms*)
  NewLine := True;
  IndDisp := 0;
  err := TStringStream.Create('');
  Syn := TStringStream.Create('');
end;

destructor TpbCocoParserGen.Destroy;
begin
  fram.Free;
  err.Free;
  Syn.Free;
  inherited;
end;


(* Put                  pbWrite ch
----------------------------------------------------------------------*)

procedure TpbCocoParserGen.Put(ch: char);
begin
  pbWrite(syn, ch)
end;

(* PutLn                pbWrite line mark
----------------------------------------------------------------------*)

procedure TpbCocoParserGen.PutLn;
begin
  pbWriteLn(syn, '')
end;

(* PutB                 pbWrite n blanks
----------------------------------------------------------------------*)

procedure TpbCocoParserGen.PutB(n: Integer); far;
begin
  if n > 0 then pbWrite(syn, StringOfChar(' ', n))
end;

(* Indent               Indent n characters
----------------------------------------------------------------------*)

procedure TpbCocoParserGen.Indent(n: Integer); far;
begin
  if NewLine then PutB(n)
  else
    NewLine := True
end;

(* IndentProc           IndentProc n characters with additional IndDisp
----------------------------------------------------------------------*)

procedure TpbCocoParserGen.IndentProc(n: Integer); far;
begin
  Indent(n + IndDisp);
end;

(* PutS                 Shortcut for WriteString(syn, ..)
----------------------------------------------------------------------*)

procedure TpbCocoParserGen.PutS(s: string); far;
var
  i: integer;
begin
  for i := 1 to Length(s) do
    if s[i] = '$' then pbWriteLn(syn, '')
  else
    pbWrite(syn, s[i]);
end;

(* PutI                 Shortcut for WriteInt(syn, i, 1)
----------------------------------------------------------------------*)

procedure TpbCocoParserGen.PutI(i: Integer);
begin
  pbWrite(syn, IntToStr(i))
end;

(* PutI2                Shortcut for WriteInt(syn, i, 2)
----------------------------------------------------------------------*)

procedure TpbCocoParserGen.PutI2(i: Integer);
begin
  pbWrite(syn, IntToStr(i))
end;

(* PutSI                Writes i or named constant of symbol i
----------------------------------------------------------------------*)

procedure TpbCocoParserGen.PutSI(i: Integer);
var
  sn: TSymbolNode;
begin
  Cocotable.GetSym(i, sn);
  if Length(sn.constant) > 0 then PutS(sn.constant)
  else
    PutI(i);
end;

(* PutSet               Enumerate bitset
----------------------------------------------------------------------*)

procedure TpbCocoParserGen.PutSet(s: BITSET; offset: Integer);
const
  MaxLine = 76;
var
  First: boolean;
  i: integer;
  l, len: integer;
  sn: TSymbolNode;
begin
  i := 0;
  First := True;
  len := 20;
  while (i < set_size) and (offset + i <= Ord(Cocotable.maxT)) do
  begin
    if i in s then
    begin
      if First then First := False
      else
      begin
        PutS(', ');
        INC(len, 2)
      end;
      Cocotable.GetSym(offset + i, sn);
      l := Length(sn.constant);
      if l > 0 then
      begin
        if len + l > MaxLine then
        begin
          PutS('$                    ');
          len := 20
        end;
        PutS(sn.constant);
        INC(len, l);
        if offset > 0 then
        begin
          Put('-');
          PutI(offset);
          INC(len, 3)
        end;
      end
      else
      begin
        if len + l > MaxLine then
        begin
          PutS('$                    ');
          len := 20
        end;
        PutI(i);
        INC(len, i div 10 + 1);
      end;
    end;
    INC(i)
  end
end;

(* PutSet1              Enumerate long set
----------------------------------------------------------------------*)

procedure TpbCocoParserGen.PutSet1(s: TCRTSet);
var
  i: integer;
  First: boolean;
begin
  i := 0;
  First := True;
  while i <= Cocotable.maxT do
  begin
    if IsIn(s, i) then
    begin
      if First then First := False
      else
        PutS(', ');
      PutSI(i)
    end;
    INC(i)
  end
end;

(* Alternatives         Count alternatives of gp
----------------------------------------------------------------------*)

function TpbCocoParserGen.Alternatives(gp: Integer): Integer;
var
  gn: TGraphNode;
  n: integer;
begin
  n := 0;
  while gp > 0 do
  begin
    Cocotable.GetNode(gp, gn);
    gp := gn.p2;
    INC(n);
  end;
  Alternatives := n;
end;

(* CopyFramePart        Copy from file <fram> to file <syn> until <stopStr>
----------------------------------------------------------------------*)

procedure TpbCocoParserGen.CopyFramePart(stopStr: string; var leftMarg: Integer);
begin
  CocoAutomaton.FrameInSize := framinSize;
  CocoAutomaton.CopyFramePart(stopStr, leftMarg, fram, syn);
end;

(* CopySourcePart       Copy sequence <pos> from input file to file <syn>
----------------------------------------------------------------------*)

procedure TpbCocoParserGen.CopySourcePart(pos: TPosition;
  indent: Integer; indentProc: TIndentProcType);
label
  999;
const
  CR = #13;
  LF = #10;
  EF = #0;
var
  lastCh, ch: char;
  extra, col, i: integer;
  bp: longint;
  nChars: longint;
begin
  if pos.beg >= 0 then
  begin
    bp := pos.beg;
    nChars := pos.len;
    col := pos.col - 1;
    ch := ' ';
    extra := 0;
    while (nChars > 0) and ((ch = ' ') or (ch = CHR(9))) do
    begin
      (* skip leading white space *)
      (* skip leading blanks *)
      Cocoscanner.CharAt(bp, ch);
      INC(bp);
      DEC(nChars);
      INC(col);
    end;
    indentProc(indent);
    while True do
    begin
      while (ch = CR) or (ch = LF) do
      begin
        (* pbWrite blank lines with the correct number of leading blanks *)
        pbWriteLn(syn, '');
        lastCh := ch;
        if nChars > 0 then
        begin
          Cocoscanner.CharAt(bp, ch);
          INC(bp);
          DEC(nChars);
        end
        else
          goto 999;
        if (ch = LF) and (lastCh = CR) then
        begin
          extra := 1 (* must be MS-DOS format *);
          if nChars > 0 then
          begin
            Cocoscanner.CharAt(bp, ch);
            INC(bp);
            DEC(nChars);
          end
          else
            EXIT;
        end;
        if (ch <> CR) and (ch <> LF) then
        (* we have something on this line *)
        begin
          indentProc(indent);
          i := col - 1 - extra;
          while ((ch = ' ') or (ch = CHR(9))) and (i > 0) do
          begin
            (* skip at most "col-1" white space chars at start of line *)
            if nChars > 0 then
            begin
              Cocoscanner.CharAt(bp, ch);
              INC(bp);
              DEC(nChars);
            end
            else
              EXIT;
            DEC(i);
          end;
        end;
      end;
      (* Handle extra blanks *)
      i := 0;
      while ch = ' ' do
      begin
        if nChars > 0 then
        begin
          Cocoscanner.CharAt(bp, ch);
          INC(bp);
          DEC(nChars)
        end
        else
          EXIT;
        INC(i);
      end;
      if (ch <> CR) and (ch <> LF) and (ch <> EF) then
      begin
        if i > 0 then PutB(i);
        pbWrite(syn, ch);
        if nChars > 0 then
        begin
          Cocoscanner.CharAt(bp, ch);
          INC(bp);
          DEC(nChars)
        end
        else
          goto 999;
      end;
    end;
    999:
  end;
end;

(* GenErrorMsg          Generate an error message and return its number
----------------------------------------------------------------------*)

procedure TpbCocoParserGen.GenErrorMsg(errTyp, errSym: Integer; var errNr: Integer);
var
  i: integer;
  Name: TsName;
  sn: TSymbolNode;
  S, S1: string;
begin
  INC(errorNr);
  errNr := errorNr;
  Cocotable.GetSym(errSym, sn);
  Name := sn.Name;
  S := StrExtract(FErrorFormat, 1, '%s');
  S1 := StrExtract(FErrorFormat, 2, '%s');
  for i := 1 to Length(Name) do
    if Name[i] = '''' then Name[i] := '"';
  pbWrite(err, ' ' + IntToStr(errNr) + ' : ' + S + '''');
  case errTyp of
    tErr: pbWrite(err, Name + ' expected');
    altErr: pbWrite(err, 'invalid ' + Name);
    syncErr: pbWrite(err, 'this symbol not expected in ' + Name);
  end;
  pbWriteLn(err, '''' + S1);
end;

(* NewCondSet    Generate a new condition set, if set not yet exists
----------------------------------------------------------------------*)

function TpbCocoParserGen.NewCondSet(newSet: TCRTSet): Integer;
var
  i: integer;
begin
  i := 1; (*skip symSet[0]*)
  while i <= maxSS do
  begin
    if Equal(newSet, symSet[i]) then
    begin
      NewCondSet := i;
      EXIT
    end;
    INC(i)
  end;
  INC(maxSS);
  if maxSS > symSetSize then Cocotable.Restriction(5, symSetSize);
  symSet[maxSS] := newSet;
  NewCondSet := maxSS
end;

(* GenCond              Generate code to check if sym is in set
----------------------------------------------------------------------*)

procedure TpbCocoParserGen.GenCond(newSet: TCRTSet; indent: Integer);
var
  i, n: integer;

  function Small(s: TCRTSet): boolean;
  begin
    i := Set_size;
    while i <= Cocotable.maxT do
    begin
      if IsIn(s, i) then
      begin
        Small := False;
        EXIT
      end;
      INC(i)
    end;
    Small := True
  end;
begin
  n := Elements(newSet, i);
  if n = 0 then PutS(' FALSE') (*this branch should never be taken*)
  else if n <= maxTerm then
  begin
    i := 0;
    while i <= Cocotable.maxT do
    begin
      if IsIn(newSet, i) then
      begin
        PutS(' (sym = ');
        PutSI(i);
        Put(')');
        DEC(n);
        if n > 0 then
        begin
          PutS(' OR');
          if Cocotable.ddt['N'] then
          begin
            PutLn;
            IndentProc(indent)
          end
        end
      end;
      INC(i)
    end
  end
  else if Small(newSet) then
  begin
    PutS(' (sym < ');
    PutI2(Set_size);
    PutS(') (* prevent range error *) AND$');
    IndentProc(indent);
    PutS(' (sym IN [');
    PutSet(newSet[0], 0);
    PutS(']) ')
  end
  else
  begin
    PutS(' _In(symSet[');
    PutI(NewCondSet(newSet));
    PutS('], sym)')
  end;
end;

(* GenCode              Generate code for graph gp in production curSy
----------------------------------------------------------------------*)

procedure TpbCocoParserGen.GenCode(gp, indent: Integer; Checked: TCRTSet);
var
  gn, gn2: TGraphNode;
  sn: TSymbolNode;
  s1, s2: TCRTSet;
  gp2, errNr, alts, indent1, addInd: integer;
  bequal: boolean;
begin
  while gp > 0 do
  begin
    Cocotable.GetNode(gp, gn);
    case gn.gnType of
      cnnonterminalSymbol:
      begin
        IndentProc(indent);
        Cocotable.GetSym(gn.p1, sn);
        PutS('_');
        PutS(sn.Name);
        if gn.pos.beg >= 0 then
        begin
          Put('(');
          NewLine := False;
          indent1 := indent + Length(sn.Name) + 2;
          CopySourcePart(gn.pos, indent1, IndentProc);
          (* was      CopySourcePart(gn.pos, 0, IndentProc); ++++ *)
          Put(')')
        end;
        PutS(';$')
      end;
      cnterminalSymbol:
      begin
        Cocotable.GetSym(gn.p1, sn);
        IndentProc(indent);
        if IsIn(Checked, gn.p1) then PutS('PerformGet;$')
        else
        begin
          PutS('Expect(');
          PutSI(gn.p1);
          PutS(');$')
        end
      end;
      cnweakTerminalSymbol:
      begin
        Cocotable.CompExpected(ABS(gn.Next), curSy, s1);
        Cocotable.GetSet(0, s2);
        Unite(s1, s2);
        Cocotable.GetSym(gn.p1, sn);
        IndentProc(indent);
        PutS('ExpectWeak(');
        PutSI(gn.p1);
        PutS(', ');
        PutI(NewCondSet(s1));
        PutS(');$')
      end;
      cnsymbolANY:
      begin
        IndentProc(indent);
        PutS('PerformGet;$')
      end;
      cnemptyAlternative:
      (* nothing *)
      begin
      end;
      cnsemanticAction:
      begin
        CopySourcePart(gn.pos, indent, IndentProc);
        //PutS(';$')
        PutS('$')
      end;
      cnsymbolSYNC:
      begin
        Cocotable.GetSet(gn.p1, s1);
        GenErrorMsg(syncErr, curSy, errNr);
        IndentProc(indent);
        PutS('WHILE NOT (');
        GenCond(s1, indent + 9);
        PutS(') DO BEGIN SynError(');
        PutI(errNr);
        PutS('); PerformGet; END;$')
      end;
      cnalternative:
      begin
        Cocotable.CompFirstSet(gp, s1);
        bequal := Equal(s1, Checked);
        alts := Alternatives(gp);
        //        OldNewLine := NewLine;
        if alts > maxAlter then
        begin
          IndentProc(indent);
          PutS('CASE sym OF$')
        end;
        gp2 := gp;
        if alts > maxAlter then addInd := 4
        else
          addInd := 2;
        //        errSemNod := -1;
        //        FirstCase := True;
        while gp2 <> 0 do
        begin
          Cocotable.GetNode(gp2, gn2);
          Cocotable.CompExpected(gn2.p1, curSy, s1);
          IndentProc(indent);
          if alts > maxAlter then
          begin
            PutS('  ');
            PutSet1(s1);
            PutS(' : BEGIN$')
          end
          else if gp2 = gp then
          begin
            PutS('IF');
            GenCond(s1, indent + 2);
            PutS(' THEN BEGIN$')
          end
          else if (gn2.p2 = 0) and bequal then
          begin
            PutS('END ELSE BEGIN$')
          end
          else
          begin
            PutS('END ELSE IF');
            GenCond(s1, indent + 5);
            PutS(' THEN BEGIN$')
          end;
          Unite(s1, Checked);
          GenCode(gn2.p1, indent + addInd, s1);
          NewLine := True;
          if alts > maxAlter then
          begin
            IndentProc(indent);
            PutS('    END;$');
          end;
          gp2 := gn2.p2;
        end;
        if not bequal then
        begin
          GenErrorMsg(altErr, curSy, errNr);
          IndentProc(indent);
          if not (alts > maxAlter) then
          begin
            PutS('END ');
          end;
          PutS('ELSE BEGIN SynError(');
          PutI(errNr);
          PutS(');$');
          if alts > maxAlter then
          begin
            IndentProc(indent);
            PutS('    END;$');
          end;
        end;
        IndentProc(indent);
        PutS('END;$');
      end;
      cniteration:
      begin
        Cocotable.GetNode(gn.p1, gn2);
        IndentProc(indent);
        PutS('WHILE');
        if gn2.gnType = cnweakTerminalSymbol then
        begin
          Cocotable.CompExpected(ABS(gn2.Next), curSy, s1);
          Cocotable.CompExpected(ABS(gn.Next), curSy, s2);
          Cocotable.GetSym(gn2.p1, sn);
          PutS(' WeakSeparator(');
          PutSI(gn2.p1);
          PutS(', ');
          PutI(NewCondSet(s1));
          PutS(', ');
          PutI(NewCondSet(s2));
          Put(')');
          Clear(s1);
          (*for inner structure*)
          if gn2.Next > 0 then gp2 := gn2.Next
          else
            gp2 := 0;
        end
        else
        begin
          gp2 := gn.p1;
          Cocotable.CompFirstSet(gp2, s1);
          GenCond(s1, indent + 5)
        end;
        PutS(' DO BEGIN$');
        GenCode(gp2, indent + 2, s1);
        IndentProc(indent);
        PutS('END;$');
      end;
      cnoption:
      begin
        Cocotable.CompFirstSet(gn.p1, s1);
        if Equal(Checked, s1) then GenCode(gn.p1, indent, Checked)
        else
        begin
          IndentProc(indent);
          PutS('IF');
          GenCond(s1, indent + 2);
          PutS(' THEN BEGIN$');
          GenCode(gn.p1, indent + 2, s1);
          IndentProc(indent);
          PutS('END;$');
        end
      end;
    end;
    if (gn.gnType <> cnemptyAlternative) and (gn.gnType <> cnsemanticAction) and
      (gn.gnType <> cnsymbolSYNC) then
      Clear(Checked);
    gp := gn.Next;
  end; (* WHILE gp > 0 *)
end;

(* GenPragmaCode        Generate code for pragmas
----------------------------------------------------------------------*)

procedure TpbCocoParserGen.GenPragmaCode(leftMarg: Integer; gramName: string);
label
  999;
var
  i: integer;
  sn: TSymbolNode;
  FirstCase: boolean;
begin
  i := Cocotable.maxT + 1;
  if i > Cocotable.maxP then EXIT;
  FirstCase := True;
  PutS('CASE sym OF$');
  PutB(leftMarg);
  while True do
  begin
    Cocotable.GetSym(i, sn);
    if FirstCase then
    begin
      FirstCase := False;
      PutS('  ')
    end
    else
    begin
      PutS('  ')
    end;
    PutSI(i);
    PutS(': BEGIN ');
    NewLine := False;
    CopySourcePart(sn.semPos, leftMarg + 6, Indent);
    PutS(' END;');
    if i = Cocotable.maxP then goto 999;
    INC(i);
    PutLn;
    PutB(leftMarg);
  end;
  999:
    PutLn;
  PutB(leftMarg);
  PutS('END;$');
  PutB(leftMarg);
  PutS('nextPos := ');
  PutS('pos;$');
  PutB(leftMarg);
  PutS('nextCol := ');
  PutS('col;$');
  PutB(leftMarg);
  PutS('nextLine := ');
  PutS('line;$');
  PutB(leftMarg);
  PutS('nextLen := ');
  PutS('len;');
end;

(* GenProcedureHeading  Generate Procedure TpbCocoParserGen.heading
----------------------------------------------------------------------*)

procedure TpbCocoParserGen.GenProcedureHeading(sn: TSymbolNode);
begin
  PutS('    Procedure ');
  PutS('_');
  PutS(sn.Name);
  if sn.attrPos.beg >= 0 then
  begin
    PutS(' (');
    NewLine := False;
    CopySourcePart(sn.attrPos, 13 + Length(sn.Name), Indent);
    Put(')')
  end;
  Put(';')
end;

(* GenForwardRefs       Generate forward references for one-pass compilers
----------------------------------------------------------------------*)

procedure TpbCocoParserGen.GenForwardRefs;
var
  sp: integer;
  sn: TSymbolNode;
begin
  sp := Cocotable.firstNt;
  while sp <= Cocotable.lastNt do
  begin (* for all nonterminals *)
    Cocotable.GetSym(sp, sn);
    GenProcedureHeading(sn);
    PutS('$');
    INC(sp)
  end;
  pbWriteLn(syn, '');
end;

(* GenProductions       Generate code for all productions
----------------------------------------------------------------------*)

procedure TpbCocoParserGen.GenProductions;
var
  sn: TSymbolNode;
  Checked: TCRTSet;
begin
  curSy := Cocotable.firstNt;
  NewLine := True; (* Bug fix PDT*)
  while curSy <= Cocotable.lastNt do
  begin (* for all nonterminals *)
    Cocotable.GetSym(curSy, sn);
    GenClassProcedureHeading(sn);
    pbWriteLn(syn, '');
    if sn.semPos.beg >= 0 then
    begin
      CopySourcePart(sn.semPos, 2, IndentProc);
      PutLn
    end;
    PutB(2);
    PutS('BEGIN$');
    {may like to add PutS(" (* "); PutS("_"); PutS(sn.name); PutS(" *)$");}
    Clear(Checked);
    GenCode(sn.struct, 4, Checked);
    PutB(2);
    PutS('END;$$');
    INC(curSy);
  end;
end;

(* GenSetInits          Initialise all sets
----------------------------------------------------------------------*)

procedure TpbCocoParserGen.InitSets;
var
  i, j: integer;
begin
  Cocotable.GetSet(0, symSet[0]);
  NewLine := False;
  i := 0;
  while i <= maxSS do
  begin
    if i <> 0 then PutLn;
    j := 0;
    while j <= Cocotable.maxT div Set_size do
    begin
      if j <> 0 then PutLn;
      Indent(2);
      PutS('symSet[');
      PutI2(i);
      PutS(', ');
      PutI(j);
      PutS('] := [');
      PutSet(symSet[i, j], j * Set_size);
      PutS('];');
      INC(j);
    end;
    INC(i)
  end
end;

(* GenCompiler          Generate the target compiler
----------------------------------------------------------------------*)

procedure TpbCocoParserGen.GenCompiler;
var
  Digits, len, pos, LeftMargin: integer;
  errNr, i, Ctr: integer;
  Checked: TCRTSet;
  gn: TGraphNode;
  sn: TSymbolNode;
  gramName, fGramName, fn, ParserFrame: string;
  temp: TStringStream;
  ch: char;
begin
  if (not CocoTable.hasParserFrame) then
    ParserFrame := Concat(Cocoscanner.directory, '\pbparser.frm')
  else 
    ParserFrame := CocoTable.ParserFrame;


  try
    Fram := LoadStringStream(ParserFrame);
    framinSize := Fram.Size;
    CocoAutomaton.FramInPos := 0;
  except
    raise Exception.Create('"pbparser.frm" not found - aborted.');
  end;

  LeftMargin := 0;
  Cocotable.GetNode(Cocotable.root, gn);
  Cocotable.GetSym(gn.p1, sn);

  gramName := sn.Name;
  ThegramName := GramName;
  //  fGramName := Concat(CocoScanner.outputDirectory, '\' + gramName);
  { TODO : Fix problem with ParserTarget }
  if (not CocoTable.hasParserTarget) then
    fGramName := Concat(CocoScanner.outputDirectory, '\' + gramName)
  else 
    fGramName := CocoTable.ParserTarget + '\' + gramName;


  fn := Concat(fGramName, '.err');
  i := 0;
  while i <= Cocotable.maxT do
  begin
    GenErrorMsg(tErr, i, errNr);
    INC(i)
  end;
  PutLn;
  PutS('// Generated by ParserBuilder @ ' + DateTimeToStr(NOW) + '$$');
  if (Cocotable.ddt['N'] or Cocotable.symNames) and not Cocotable.ddt['D'] then
  begin
    PutS('UNIT ');
    PutS(gramName);
    PutS('G;$$');
    PutS('INTERFACE$$');
    //    PutLn;
    //    PutS('// Generated by ParserBuilder @ '+DateTimeToStr(NOW)+'$$');
    PutS('CONST');
    i := 0;
    pos := MaxSourceLineLength + 1;
    repeat
      Cocotable.GetSym(i, sn);
      len := Length(sn.constant);
      if len > 0 then
      begin
        errNr := i;
        Digits := 1;
        while errNr >= 10 do
        begin
          INC(Digits);
          errNr := errNr div 10
        end;
        INC(len, 3 + Digits + 1);
        if pos + len > MaxSourceLineLength then
        begin
          PutLn;
          pos := 0
        end;
        PutS('  ');
        PutS(sn.constant);
        PutS(' = ');
        PutI(i);
        Put(';');
        INC(pos, len + 2);
      end;
      INC(i);
    until i > Cocotable.maxP;
    PutS('$$IMPLEMENTATION$');
    PutS('END.$');
    fn := Concat(fGramName, 'g.pas');

    if FileExists(fn) then
      CopyFile(PChar(fn), PChar(ChangeFileExt(fn, '.pas.backup')), False);

    SaveStringStream(Syn, fn);
  end;

  (* IF Cocotable.ddt["N"] OR Cocotable.symNames *)
  (*----- pbWrite *p.pas -----*)
  Syn.Free;
  Syn := TStringStream.Create('');
  PutS('// Generated by ParserBuilder @ ' + DateTimeToStr(NOW) + '$$');
  CopyFramePart('-->modulename', LeftMargin);
  PutS(gramName);

  CopyFramePart('-->UsedUnits', LeftMargin);

  // Writes out the uses clause
  if Cocotable.hasUses then
  begin
    CopySourcePart(Cocotable.useDeclPos, 0, PutB);
    PutS(', ')
  end;

  PutS(gramName + 'Scanner');

  if Cocotable.ddt['N'] or Cocotable.symNames then
    CoCoAutomaton.ImportSymConsts(', ', PutS)
  else
    PutS(';$');
  // End of uses clause

  CopyFramePart('-->constants', LeftMargin);
  PutS('maxT = ');
  PutI(Cocotable.maxT);
  Put(';');
  if Cocotable.maxP > Cocotable.maxT then
  begin
    PutLn;
    PutB(LeftMargin);
    PutS('maxP = ');
    PutI(Cocotable.maxP);
    Put(';')
  end;
  // New Terms for the ATG file
  // GlobalDecs are those that come before the generated Class Definition
  CopyFramePart('-->GlobalDecs', LeftMargin);
  if Cocotable.hasGlobalDecs then
  begin
    CopySourcePart(Cocotable.globalDeclPos, 0, PutB);
  end;

  CopyFramePart('-->modulename', LeftMargin);
  PutS(gramName);

  CopyFramePart('-->modulename', LeftMargin);
  PutS(gramName);


  CopyFramePart('-->PrivateDecs', LeftMargin);
  if Cocotable.hasPrivateDecs then
  begin
    CopySourcePart(Cocotable.PrivateDeclPos, 0, PutB);
  end;

  CopyFramePart('-->Classdeclarations', LeftMargin);
  GenForwardRefs;

  CopyFramePart('-->ProtectedDecs', LeftMargin);
  if Cocotable.hasProtectedDecs then
  begin
    CopySourcePart(Cocotable.ProtectedDeclPos, 0, PutB);
  end;

  CopyFramePart('-->PublicDecs', LeftMargin);
  if Cocotable.hasPublicDecs then
  begin
    CopySourcePart(Cocotable.PublicDeclPos, 0, PutB);
  end;

  CopyFramePart('-->PublishedDecs', LeftMargin);
  if Cocotable.hasPublishedDecs then
  begin
    CopySourcePart(Cocotable.PublishedDeclPos, 0, PutB);
  end;

  // End of New Terms for the ATG file

  CopyFramePart('-->symSetSize', LeftMargin);
  pbWrite(syn, chr(255)); (* marker *)

  CopyFramePart('-->ArbitrarySource', LeftMargin);
  CopySourcePart(Cocotable.semDeclPos, 0, PutB);

  CopyFramePart('-->modulename', LeftMargin);
  PutS(gramName);
  CopyFramePart('-->modulename', LeftMargin);
  PutS(gramName);


  CopyFramePart('-->error', LeftMargin);
  PutS('Error(errNo, ');
  PutS('line, ');
  PutS('col, ');
  PutS('pos);');

  CopyFramePart('-->modulename', LeftMargin);
  PutS(gramName);

  CopyFramePart('-->error', LeftMargin);
  PutS('Error(errNo, ');
  PutS('nextLine, ');
  PutS('nextCol, ');
  PutS('nextPos);');
  CopyFramePart('-->modulename', LeftMargin);
  PutS(gramName);

  CopyFramePart('-->pragmas', LeftMargin);
  GenPragmaCode(LeftMargin, gramName);

  for i := 1 to 10 do
  begin
    CopyFramePart('-->modulename', LeftMargin);
    PutS(gramName);
  end;

  CopyFramePart('-->productions', LeftMargin);
  GenProductions;

  CopyFramePart('-->modulename', LeftMargin);
  PutS(gramName);
  CopyFramePart('-->parseRoot', LeftMargin);

  PutS('_Reset; PerformGet;$');

  Clear(Checked);
  GenCode(Cocotable.root, LeftMargin, Checked);

  CopyFramePart('-->modulename', LeftMargin);
  PutS(gramName);
  CopyFramePart('-->Destroy', LeftMargin);
  if Cocotable.hasDestroyDecs then
  begin
    CopySourcePart(Cocotable.DestroyDeclPos, 0, PutB);
  end;


  CopyFramePart('-->modulename', LeftMargin);
  PutS(gramName);

  CopyFramePart('-->Create', LeftMargin);
  if Cocotable.hasCreateDecs then
  begin
    CopySourcePart(Cocotable.CreateDeclPos, 0, PutB);
  end;

  CopyFramePart('-->initialization', LeftMargin);
  InitSets;
  CopyFramePart('-->modulename', LeftMargin);
  PutS(gramName + 'Parser *)$');

  //  fn := Concat(fGramName, 'P.$$$');
  fn := ChangeFileExt(fGramName, 'P.$$$');
  SaveStringStream(Syn, fn);

  fn := ChangeFileExt(fGramName, '.err');
  //  fn := Concat(fGramName, '.err');

  if FileExists(fn) then
    CopyFile(PChar(fn), PChar(ChangeFileExt(fn, '.err.backup')), False);
  if Cocotable.hasErrorsSource then
    pbWriteLn(err, Cocotable.ErrorsSource);
  SaveStringStream(err, fn);
  if maxSS < 0 then maxSS := 0;
  Syn.Free;
  Syn := TstringStream.Create('');
  fn := Concat(fGramName, 'P.$$$');
  if (FileExists(fn)) then
  begin
    Temp := LoadStringStream(fn);
    framinSize := Temp.Size;
  end
  else
    raise Exception.Create('Cannot find temp file - ' + fn);

  Ctr := 0;
  if (Temp <> nil) then
  begin
    while Ctr < framinSize do
    begin
      pbRead(Temp, ch);
      if ch = CHR(255) then pbWrite(syn, IntToStr(maxSS))
      else
        pbWrite(syn, ch);
      inc(Ctr);
    end;
    Temp.Free;
  end;

  fn := fGramName + 'Parser.pas';
  if FileExists(fn) then
    CopyFile(PChar(fn), PChar(ChangeFileExt(fn, '.pas.backup')), False);

  SaveStringStream(Syn, fn);
  DeleteFile(concat(fGramName, 'P.$$$'));
end;

(* WriteStatistics      Write statistics about compilation to list file
----------------------------------------------------------------------*)

procedure TpbCocoParserGen.WriteStatistics;

  procedure WriteNumbers(used, available: Integer);
  begin
    pbWriteLn(CocoScanner.lst, IntToStr(used + 1) + ' (limit ' + IntToStr(available) + ')');
  end;
begin
  pbWriteLn(CocoScanner.lst, 'Statistics:');
  pbWriteLn(CocoScanner.lst, '');
  pbWrite(CocoScanner.lst, '  nr of terminals:    ');
  WriteNumbers(Cocotable.maxT, maxTerminals);
  pbWrite(CocoScanner.lst, '  nr of non-terminals:');
  WriteNumbers(Cocotable.lastNt - Cocotable.firstNt,
    maxSymbols - maxT - 1);
  pbWrite(CocoScanner.lst, '  nr of pragmas:      ');
  WriteNumbers(maxSymbols - Cocotable.lastNt - 2,
    maxSymbols - maxT - 1);
  pbWrite(CocoScanner.lst, '  nr of symbolnodes:  ');
  WriteNumbers(maxSymbols - Cocotable.firstNt + Cocotable.maxT,
    maxSymbols);
  pbWrite(CocoScanner.lst, '  nr of graphnodes:   ');
  WriteNumbers(Cocotable.nNodes, maxNodes);
  pbWrite(CocoScanner.lst, '  nr of conditionsets:');
  WriteNumbers(maxSS, symSetSize);
  pbWrite(CocoScanner.lst, '  nr of charactersets:');
  WriteNumbers(Cocotable.maxC, maxClasses);
  pbWriteLn(CocoScanner.lst, '');
  pbWriteLn(CocoScanner.lst, '');
end;

procedure TpbCocoParserGen.SetCocoAutomaton(const Value: TpbCocoAutomaton);
begin
  FCocoAutomaton := Value;
end;

procedure TpbCocoParserGen.SetCocoScanner(const Value: TpbCocoScanner);
begin
  FCocoScanner := Value;
end;

procedure TpbCocoParserGen.SetCocoTable(const Value: TpbCocoTable);
begin
  FCocoTable := Value;
end;

procedure TpbCocoParserGen.GenClassProcedureHeading(sn: TSymbolNode);
begin
  PutS('Procedure T' + ThegramName + 'Parser.');
  PutS('_');
  PutS(sn.Name);
  if sn.attrPos.beg >= 0 then
  begin
    PutS(' (');
    NewLine := False;
    CopySourcePart(sn.attrPos, 13 + Length(sn.Name), Indent);
    Put(')')
  end;
  Put(';')
end;

procedure TpbCocoParserGen.SetErrorFormat(const Value: string);
begin
  FErrorFormat := Value;
end;

end.
