{-----------------------------------------------------------------------------
 Unit Name: TreeEditSync
 Author:    Pat Connors
 Purpose:
         Takes responsibility for synchronising the edits taking place within
         the GUI, especially updating the tree.

-----------------------------------------------------------------------------}


unit TreeEditSync;

interface

uses
  Windows, Messages, SysUtils, Classes, pbStrLib, ComCtrls,
  CompilerObj;

type
  TTreeEditSync = class(TComponent)
  private
    FTreeView: TTreeView;
    FCompiler: TCompiler;
    FEditorStyle: Integer;
    FLSNodeCaption: string;
    FLSNodeType: TCompilerObjectType;
    FLSNodeIndex: Integer;
    procedure SetTreeView(const Value: TTreeView);
    function GetNodeText: string;
    procedure SetCompiler(const Value: TCompiler);
    procedure SetEditorStyle(const Value: Integer);
    procedure SetLSNodeCaption(const Value: string);
    procedure SetLSNodeType(const Value: TCompilerObjectType);
    procedure SetLSNodeIndex(const Value: Integer);
  protected
    { Protected declarations }
    function BranchOfCompiler(TN: TTreeNode): boolean;
  public
    { Public declarations }
    function getCocoEditorType: Integer;
    function getCocoEditorText: string;
    function GetSelectedCompiler: TCompiler;
    procedure ReselectNode;
    procedure SelectCompilerRootNode;
    procedure SelectNodeNamed(const S: string);
    procedure EditCurrentNode;
    function GetEditorStyleForNode(const NodeText: string): Integer;
  published
    { Published declarations }
    property TreeView: TTreeView read FTreeView write SetTreeView;
    property Compiler: TCompiler read FCompiler write SetCompiler;
    property EditorStyle: Integer read FEditorStyle write SetEditorStyle;
    property LSNodeCaption: string read FLSNodeCaption write SetLSNodeCaption;
    property LSNodeIndex: Integer read FLSNodeIndex write SetLSNodeIndex;
    property LSNodeType: TCompilerObjectType read FLSNodeType write SetLSNodeType;
  end;


implementation

uses CocoComments, DelphiDeclarations;


{ TTreeEditSync }

{-----------------------------------------------------------------------------
  Procedure: TTreeEditSync.BranchOfCompiler
  Arguments: TN: TTreeNode
  Result:    Boolean = True if the branch holding the node TN is that of
                       the current compiler.
-----------------------------------------------------------------------------}
function TTreeEditSync.BranchOfCompiler(TN: TTreeNode): boolean;
begin
  Result := False;
  // default position = better to not select a node than to incorrectly select one
  if TN.Data <> nil then
  begin
    Result := (TN.Data = FCompiler);
    Exit;
  end
  else
    while (TN <> TreeView.Items[0]) and (TN.Parent <> nil) do
    begin
      TN := TN.Parent;
      if TN.Data <> nil then
      begin
        Result := (TN.Data = FCompiler);
        Exit;
      end
    end;
end;

function TTreeEditSync.getCocoEditorText: string;
var
  S, NodeCaption: string;
  NodeIdx: integer;
begin
  FEditorStyle := 0;
  if (fCompiler = nil) then
  begin
    Result := '';
    Exit;
  end;
  FCompiler.ResetAllEditFlags;
  NodeCaption := GetNodeText;
  if (NodeCaption = 'Delphi') then
  begin
    Result := FCompiler.DelphiSection.GetATG;
    FCompiler.EditCompilerObject(coDelphi, '');
  end
  else if (NodeCaption = 'Uses') then
  begin
    Result := FCompiler.DelphiSection.UsesClause;
    FCompiler.EditCompilerObject(coUses, '');
  end
  else if (NodeCaption = 'Global') then
  begin
    Result := FCompiler.DelphiSection.TypeDeclarations;
    FCompiler.EditCompilerObject(coGlobal, '');
  end
  else if (NodeCaption = 'Private') then
  begin
    Result := FCompiler.DelphiSection.PrivateDecs;
    FCompiler.EditCompilerObject(coPrivate, '');
  end
  else if (NodeCaption = 'Errors') then
  begin
    Result := FCompiler.DelphiSection.ErrorsSource;
    FCompiler.EditCompilerObject(coErrors, '');
  end
  else if (NodeCaption = 'Protected') then
  begin
    Result := FCompiler.DelphiSection.ProtectedDecs;
    FCompiler.EditCompilerObject(coProtected, '');
  end
  else if (NodeCaption = 'Public') then
  begin
    Result := FCompiler.DelphiSection.PublicDecs;
    FCompiler.EditCompilerObject(coPublic, '');
  end
  else if (NodeCaption = 'Published') then
  begin
    Result := FCompiler.DelphiSection.PublishedDecs;
    FCompiler.EditCompilerObject(coPublished, '');
  end
  else if (NodeCaption = 'Create') then
  begin
    Result := FCompiler.DelphiSection.CreateSrc;
    FCompiler.EditCompilerObject(coCreate, '');
  end
  else if (NodeCaption = 'Destroy') then
  begin
    Result := FCompiler.DelphiSection.DestroySrc;
    FCompiler.EditCompilerObject(coDestroy, '');
  end
  else if (NodeCaption = 'Arbitrary Code') then
  begin
    Result := FCompiler.DelphiSection.ArbitraryCodeSection;
    FCompiler.EditCompilerObject(coArbitraryCode, '');
  end
  else if (Pos(fCompiler.CompilerName + ' :', NodeCaption) = 1) then
  begin
    Result := FCompiler.GetATG;
    FCompiler.EditCompilerObject(coCompiler, '');
  end
  else if (NodeCaption = 'Options') then
  begin
    Result := '';
    FEditorStyle := 1; // 1 = Options
  end
  else if (NodeCaption = 'Ignore') then
  begin
    Result := FCompiler.IgnoreSection.GetATG;
    FCompiler.EditCompilerObject(coIgnore, '');
  end
  else if (NodeCaption = 'Pragmas') then
  begin
    Result := FCompiler.PragmasSection.GetATG;
    FCompiler.EditCompilerObject(coPragmas, '');
  end
  else if (NodeCaption = 'Names') then
  begin
    Result := 'NAMES' + CRLF + FCompiler.Names;
    FCompiler.EditCompilerObject(coNames, '');
  end
  else if (NodeCaption = 'Productions') then
  begin
    FCompiler.Productions.saMinIndent := fCOmpiler.saMinIndent;
    Result := FCompiler.Productions.GetATG;
    FCompiler.EditCompilerObject(coProductions, '');
  end
  else if (NodeCaption = 'Tokens') then
  begin
    Result := FCompiler.TokenSection.GetATG;
    FCompiler.EditCompilerObject(coTokens, '');
  end
  else if (NodeCaption = 'Characters') then
  begin
    Result := FCompiler.CharactersSection.GetATG;
    FCompiler.EditCompilerObject(coCharacters, '');
  end
  else if (NodeCaption = 'Comments') then
  begin
    Result := FCompiler.CocoComments.GetATG;
    FCompiler.EditCompilerObject(coComments, '');
  end
  else if (Pos('P:', NodeCaption) = 1) then
  begin
    S := Copy(NodeCaption, 3, Length(NodeCaption));
    FCompiler.Productions.saMinIndent := fCOmpiler.saMinIndent;
    Result := FCompiler.Productions.GetProductionTitled(S).GetATG;
    FEditorStyle := 0;
    FCompiler.EditCompilerObject(coProductionNode, S);
  end
  else if (Pos('Path ', NodeCaption) = 1) then
  begin
    // Get the index of the item
    S := StrExtract(NodeCaption, 1, ':');
    S := StrExtract(S, 2, ' ');
    NodeIdx := StrToInt(S);
    S := StrExtract(NodeCaption, 2, ':');
    Result := FCompiler.Productions.GetProductionTitled(S).GetPath(NodeIdx).GetATG;
    { TODO : Add editor flags for production paths }
    //    FCompiler.EditCompilerObject(coProductionNode,S);
  end
  //
  else if (Pos('CHAR:', NodeCaption) = 1) then
  begin
    S := StrExtract(NodeCaption, 2, ':');
    Result := FCompiler.CharactersSection.GetCharacterSet(S);
    FCompiler.EditCompilerObject(coCharacterNode, S);
  end
  else if (Pos('TOK:', NodeCaption) = 1) then
  begin
    S := StrExtract(NodeCaption, 2, ':');
    Result := FCompiler.TokenSection.GetTokenDefinition(S).GetATG;
    FCompiler.EditCompilerObject(coTokenNode, S);
  end
  else if (Pos('PRA:', NodeCaption) = 1) then
  begin
    S := StrExtract(NodeCaption, 2, ':');
    Result := FCompiler.PragmasSection.GetPragmaDefinition(S).GetATG;
    FCompiler.EditCompilerObject(coPragmasNode, S);
  end
  else if (Pos('PRI:', NodeCaption) = 1) or
    (Pos('PRO:', NodeCaption) = 1) or
    (Pos('PBC:', NodeCaption) = 1) or
    (Pos('PBS:', NodeCaption) = 1) then
  begin
    { TODO : What to do about editing the Delphi Source Code? }
    S := StrExtract(NodeCaption, 2, ':');
    Result := S;
  end
  else if (Pos(FCompiler.Filename, NodeCaption) > 0) then
  begin
    S := FCompiler.GetATG;
    Result := S;
  end
  else
    Result := NodeCaption;
  //  else
  //    Raise Exception.Create('NodeCaption = ' + NodeCaption);
  { TODO : Add other node titles here }
end;

function TTreeEditSync.getCocoEditorType: Integer;
begin
  Result := 0; // Text
end;

procedure TTreeEditSync.EditCurrentNode;
var 
  S: string;
begin
  if (Treeview.Selected <> nil) then
  begin
    S := Treeview.Selected.Text;
    if (S = 'Delphi') then
      FCompiler.EditCompilerObject(coDelphi, '')
    else if (S = 'Uses') then
      FCompiler.EditCompilerObject(coUses, '')
    else if (S = 'Global') then
      FCompiler.EditCompilerObject(coGlobal, '')
    else if (S = 'Private') then
      FCompiler.EditCompilerObject(coPrivate, '')
    else if (S = 'Protected') then
      FCompiler.EditCompilerObject(coProtected, '')
    else if (S = 'Public') then
      FCompiler.EditCompilerObject(coPublic, '')
    else if (S = 'Published') then
      FCompiler.EditCompilerObject(coPublished, '')
    else if (S = 'Create') then
      FCompiler.EditCompilerObject(coCreate, '')
    else if (S = 'Destroy') then
      FCompiler.EditCompilerObject(coDestroy, '')
    else if (S = 'Arbitrary Code') then
      FCompiler.EditCompilerObject(coArbitraryCode, '')
    else if (Pos(fCompiler.CompilerName + ' :', S) = 1) then
      FCompiler.EditCompilerObject(coCompiler, '')
    else if (S = 'Options') then
      FCompiler.EditCompilerObject(coOptions, '')
    else if (S = 'Ignore') then
      FCompiler.EditCompilerObject(coIgnore, '')
    else if (S = 'Pragmas') then
      FCompiler.EditCompilerObject(coPragmas, '')
    else if (S = 'Names') then
      FCompiler.EditCompilerObject(coNames, '')
    else if (S = 'Productions') then
      FCompiler.EditCompilerObject(coProductions, '')
    else if (S = 'Tokens') then
      FCompiler.EditCompilerObject(coTokens, '')
    else if (S = 'Characters') then
      FCompiler.EditCompilerObject(coCharacters, '')
    else if (S = 'Comments') then
      FCompiler.EditCompilerObject(coComments, '')
    else if (Treeview.Selected.Parent <> nil) then
      if Treeview.Selected.Parent.Text = 'Productions' then
        FCompiler.EditCompilerObject(coProductionNode, S)
    end;
end;

function TTreeEditSync.GetEditorStyleForNode(const NodeText: string): Integer;
begin
  if (NodeText = 'Frames') then Result := 2
  else if (NodeText = 'Options') then Result := 1
  else 
    Result := 0;
end;

function TTreeEditSync.GetNodeText: string;
var
  p: Pointer;
begin
  Result := '';
  //  FEditorStyle := 0;
  if (Treeview.Selected <> nil) then
    p := Treeview.Selected.Data
  else
    Exit;
  FLSNodeCaption := '';
  fLSNodeIndex := -1;
  if (P <> nil) then FCompiler := TCompiler(P);
  FLSNodeCaption := Treeview.Selected.Text;

  if (P <> nil) or (Treeview.Selected.Text = 'Options') then
  begin
    Result := Treeview.Selected.Text;
  end
  else if (Treeview.Selected.Text = 'Untouched ATG') then
  begin
    Result := FCompiler.OriginalATG;
  end
  else if (Treeview.Selected.Text = 'Frames') then
  begin
    Result := '';
    FEditorStyle := 2;
  end
  else if (Treeview.Selected.Text = 'Comments') then
  begin
    Result := 'Comments';
    FEditorStyle := 0;
  end
  else if (Treeview.Selected.Text = 'Header') then
  begin
    Result := Trim(fCompiler.Header);
    FEditorStyle := 0;
    FCompiler.EditCompilerObject(coHeader, '');
  end
  else if (Treeview.Selected.Text = 'Change History') then
  begin
    Result := '';
    FEditorStyle := 5;
  end
  else if (Treeview.Selected.Text = 'Targets') then
  begin
    Result := '';
    FEditorStyle := 2;
  end
  // resume previously coded processing
  else
  begin
    if (Treeview.Selected.Parent <> nil) then
    begin
      Result := Treeview.Selected.Parent.Text;
      if (Result = 'Characters') then
      begin
        Result := 'CHAR:' + Treeview.Selected.Text;
      end
      else if (Result = 'Tokens') then
      begin
        Result := 'TOK:' + Treeview.Selected.Text;
      end
      else if (Result = 'Pragmas') then
      begin
        Result := 'PRA:' + Treeview.Selected.Text;
      end
      else if (Result = 'Private') then
      begin
        Result := 'PRI:' + Treeview.Selected.Text;
      end
      else if (Result = 'Protected') then
      begin
        Result := 'PRO:' + Treeview.Selected.Text;
      end
      else if (Result = 'Public') then
      begin
        Result := 'PBC:' + Treeview.Selected.Text;
      end
      else if (Result = 'Published') then
      begin
        Result := 'PBS:' + Treeview.Selected.Text;
      end
      else
      begin
        P := Treeview.Selected.Parent.Data;
        if (p <> nil) then
        begin
          // if the Data <> NIL then the selectnode must be a Production
          Result := 'P:' + Treeview.Selected.Text;
          FLSNodeIndex := Treeview.Selected.Parent.IndexOf(Treeview.Selected);
        end
        else
        begin
          // if the Data = NIL then the selectnode must be a Single Production Path
          Result := Treeview.Selected.Text + ':' + Treeview.Selected.Parent.Text;
        end;
      end;
    end;
  end;
end;

function TTreeEditSync.GetSelectedCompiler: TCompiler;
var
  p: Pointer;
  TN: TTreeNode;
  TopLevelNodeCount, TopLevelNodeIdx, ChildNodeIdx, SubChildNodeIdx: integer;
begin
  //  Assert(Treeview.Selected <> nil, 'BUG: No node selected - should catch this one upstream');
  Result := nil;
  // Modification 19-05-02 - pc
  // Previously
  //  if (Treeview.Selected = nil) then
  //    Exit;
  //  ... but it wasn't good - as we should never reach this point unless
  // there is a compiler loaded. If there is no selected node, it is due to
  // something else (out of control GUI updating or non-updating).
  // Will try seeking out the most obvious compiler (if multiples exist) based on the
  // presence of a node with the same title as the last selected node.
  // If that name is blank, will simply choose the first compiler in the tree.
  if Treeview.Selected = nil then
  begin
    if (LSNodeCaption = '') then
    begin
      if TreeView.Items[0] <> nil then
        if TreeView.Items[0].Count > 0 then
        begin
          Result := TCompiler(TreeView.Items[0].Item[0].Data);
          Exit;
        end;
    end
    else
    begin
      // LSNodeCaption <> ''
      if TreeView.Items[0] <> nil then
        if TreeView.Items[0].Count > 0 then
        begin
          TopLevelNodeCount := TreeView.Items[0].Count;
          for TopLevelNodeIdx := 0 to TopLevelNodeCount - 1 do
          begin
            TN := TreeView.Items[0].Item[TopLevelNodeIdx];
            p := TN.Data;
            if TN.Text = LSNodeCaption then
            begin
              Result := p;
              Exit;
            end
            else
            begin
              for ChildNodeIdx := 0 to TN.Count - 1 do
              begin
                if TN.Item[ChildNodeIdx].Text = LSNodeCaption then
                begin
                  Result := p;
                  Exit;
                end
                else
                begin
                  // Probe the children of this node - this will be as deep as the
                  // tree goes

                  for SubChildNodeIdx := 0 to TN.Item[ChildNodeIdx].Count - 1 do
                  begin
                    if TN.Item[ChildNodeIdx].Item[SubChildNodeIdx].Text = LSNodeCaption then
                    begin
                      Result := p;
                      Exit;
                    end;
                  end;
                end;
              end;
            end;
          end;
        end;
    end;
  end;




  if (Treeview.Selected <> TreeView.Items[0]) then
  begin
    p := Treeview.Selected.Data;
    if (P <> nil) then Result := TCompiler(P)
    else
    begin
      // Walk back up the tree to see which compiler (first Data<> NIL encountered) owns this node
      TN := Treeview.Selected;
      while (TN.Parent <> TreeView.Items[0]) and (TN.Data = nil) do
        TN := TN.Parent;
      Result := TCompiler(TN.Data);
      if (Result = nil) and (FCompiler <> nil) then Result := FCompiler;
      Assert(Result <> nil, 'Result cannot point to a NULL compiler.');
    end;
  end;
end;

procedure TTreeEditSync.ReselectNode;
var
  idx: integer;
begin
  //  Assert(FLSNodeCaption <> '');
  if FLSNodeCaption <> '' then
    for idx := 0 to TreeView.Items.Count - 1 do
    begin
      if (Treeview.Items[idx].Text = FLSNodeCaption) and
        (BranchOfCompiler(Treeview.Items[idx])) then
      begin
        Treeview.Select([Treeview.Items[idx]]);
        Exit;
      end;
    end;

  // If we are still in this method, then the user
  // has changed the productions - either changed the name(s) or
  // has deleted a production.
  // We have the index of the last selected production if
  // fLSNodeIndex>-1.
  // if it is smaller than the production count, we will
  // select that node, else we will select the all-inclusive Productions node.
  if (fLSNodeIndex>-1) then
  begin
    if (fLSNodeIndex < FCompiler.Productions.Count) then
      FLSNodeCaption := FCompiler.Productions.GetProduction(flsNodeIndex).ProductionName
    else
      FLSNodeCaption := 'Productions';
    // so we only recurse the once
    fLSNodeIndex := -1;
    ReselectNode;
  end;
end;

procedure TTreeEditSync.SelectCompilerRootNode;
var 
  Tn: TTReeNode;
  idx, ChildCnt: integer;
begin
  if FTreeView.Items[0] <> nil then
  begin
    ChildCnt := FTreeView.Items[0].Count;
    for idx := 0 to ChildCnt - 1 do
    begin
      TN := FTreeView.Items[0].Item[idx];
      if TN.Data = FCompiler then
      begin
        fTreeView.Selected := TN;
      end
      else
        TN.Collapse(True);
    end;
  end;
end;

procedure TTreeEditSync.SelectNodeNamed(const S: string);
var 
  idx: integer;
  R: TRect;
begin
  Assert(FLSNodeCaption <> '');
  for idx := 0 to TreeView.Items.Count - 1 do
  begin
    if (UpperCase(Treeview.Items[idx].Text) = UpperCase(S)) and
      (BranchOfCompiler(Treeview.Items[idx])) then
    begin
      Treeview.Items[idx].MakeVisible;
      R := Treeview.Items[idx].DisplayRect(False);
      if (R.Top < TreeView.Top + 30) then
        TreeView.Perform(WM_VSCROLL, SB_LINEDOWN, 5)
      else if (R.Top > (TreeView.Top + TreeView.Height - 30)) then
        TreeView.ScrollBy(0, - 3);
      Compiler.ResetAllEditFlags;
      Treeview.Select([Treeview.Items[idx]]);
      FLSNodeCaption := Treeview.Items[idx].Text;
      Exit;
    end;
  end;
end;

procedure TTreeEditSync.SetCompiler(const Value: TCompiler);
begin
  FCompiler := Value;
end;

procedure TTreeEditSync.SetEditorStyle(const Value: Integer);
begin
  FEditorStyle := Value;
end;

procedure TTreeEditSync.SetLSNodeCaption(const Value: string);
begin
  FLSNodeCaption := Value;
end;

procedure TTreeEditSync.SetLSNodeType(const Value: TCompilerObjectType);
begin
  FLSNodeType := Value;
end;

procedure TTreeEditSync.SetTreeView(const Value: TTreeView);
begin
  FTreeView := Value;
end;

procedure TTreeEditSync.SetLSNodeIndex(const Value: Integer);
begin
  FLSNodeIndex := Value;
end;

end.
